from __future__ import annotations

import logging
from collections import namedtuple

import numpy
from silx.io.url import DataUrl
from silx.utils.enum import Enum as _Enum
from tomoscan.esrf.scan.utils import get_data

from tomwer.core.process.reconstruction.utils.cor import relative_pos_to_absolute
from tomwer.core.process.reconstruction.axis.side import Side
from tomwer.core.scan.scanbase import TomwerScanBase

from .anglemode import CorAngleMode
from .mode import AxisMode, AXIS_MODE_METADATAS
from .projectiontype import ProjectionType

from nabu.preproc.phase import PaganinPhaseRetrieval


_logger = logging.getLogger(__name__)


_calculation_conf = namedtuple(
    "_calculation_conf", ["projection_type", "paganin"]
)  # noqa

_WITH_PAG = "withpag"
_NO_PAG = "nopag"


DEFAULT_CMP_THETA = 5
DEFAULT_CMP_N_SUBSAMPLING_Y = 40
DEFAULT_CMP_NEAR_POS = 0
DEFAULT_CMP_NEAR_WIDTH = 40
DEFAULT_CMP_OVERSAMPLING = 4
DEFAULT_CMP_TAKE_LOG = True


class AxisCalculationInput(_Enum):
    """Define the different mode of input the user can have for axis calculation"""

    emission = _calculation_conf(ProjectionType.absorption, False)
    transmission = _calculation_conf(ProjectionType.transmission, False)
    transmission_pag = _calculation_conf(ProjectionType.transmission, True)

    def name(self):  # pylint: disable=E0102
        if self.value.paganin is True:
            return " ".join((self.value.projection_type.value, "paganin"))
        else:
            return self.value.projection_type.value

    def to_dict(self):
        pag_text = _WITH_PAG if self.value.paganin is True else _NO_PAG
        return "_".join((self.value.projection_type.value, pag_text))

    @classmethod
    def from_value(cls, value):
        if type(value) is str and len(value.split("_")) == 2:
            proj_type, pag = value.split("_")
            value_pag = True if pag == _WITH_PAG else False
            value_proj = ProjectionType.from_value(proj_type)
            return cls.from_value(_calculation_conf(value_proj, value_pag))
        elif isinstance(value, cls):
            return value
        else:
            for member in cls:
                if value in (member.value, member.name()):
                    return member
            raise ValueError(f"Cannot convert: {value}")


class AxisResource(object):
    """Helper for axis file relative stuff"""

    _PAGANIN_CONFIG = {"distance": 100e-3, "energy": 35, "delta_beta": 1e3}
    """Paganin configuration for axis calculation. To simplify we have only one
    static configuration for now. Otherwise complicate stuff"""

    def __init__(self, url: None | DataUrl, angle: float | None = None):
        assert url is None or isinstance(url, DataUrl)
        assert url is None or url.is_valid()
        self.__url = url
        self.__raw_data = None
        self.__norme_data = None
        self.__norm_paganin = None
        self.__angle = angle

    def __str__(self):
        return f"{type(self)}, url: {self.__url.path() if self.__url else None}"

    @property
    def url(self):
        return self.__url

    @property
    def data(self):
        """

        :return: 2D numpy.array
        """
        if self.__url is None:
            return None
        if self.__raw_data is None:
            self.__raw_data = get_data(self.url)
        return self.__raw_data

    @data.setter
    def data(self, data):
        self.__raw_data = data

    @property
    def normalized_data(self):
        return self.__norme_data

    @normalized_data.setter
    def normalized_data(self, data):
        self.__norme_data = data

    @property
    def angle(self) -> float | None:
        return self.__angle

    def normalize_data(self, scan, log_):
        """Normalize data for axis calculation"""
        if not isinstance(scan, TomwerScanBase):
            raise TypeError(
                f"scan is expected to be an instance of {TomwerScanBase}. {type(scan)} provided"
            )
        if self.__url is None:
            return None
        else:
            self.__norme_data = scan.data_flat_field_correction(data=self.data)
            if log_ is True:
                self.__norme_data[numpy.isnan(self.__norme_data)] = (
                    self.__norme_data.max()
                )
                if self.__norme_data.max() < 0:
                    _logger.error("max data value < 0 unable to compute log")
                else:
                    try:
                        data = self.__norme_data
                        data[data <= 1] = 1
                        self.__norme_data = -numpy.log(data)
                    except Exception as e:
                        _logger.error("Fail to apply log on radio" + e)
            else:
                self.__norme_data[numpy.isnan(self.__norme_data)] = (
                    self.__norme_data.min()
                )

        return self.__norme_data

    @property
    def normalized_data_paganin(self):
        """

        :return: data processed by the Pagagin phase retrieval
        """
        if self.__url is None:
            return None
        if self.__norm_paganin is None:
            if self.normalized_data is None:
                raise ValueError("data should be normalized before applying " "paganin")
            else:
                data = self.normalized_data
                phase_retrieval = PaganinPhaseRetrieval(
                    data.shape, **AxisResource._PAGANIN_CONFIG
                )
                self.__norm_paganin = phase_retrieval.apply_filter(data)
        return self.__norm_paganin

    @normalized_data_paganin.setter
    def normalized_data_paganin(self, data):
        self.__norm_paganin = data

    def __eq__(self, other):
        if not isinstance(other, AxisResource):
            return False
        if self.url is None and other.url is None:
            return True
        elif self.url is None or other.url is None:
            return False
        else:
            return self.url == other.url


class AxisRP:
    """
    Configuration class for a tomwer :class:`AxisTask`

    note: every modification on the parameters will process a call fo changed
    except `axis_url_1` and `axis_url_2` which will produce a call to the
    dedicated axis_url_changed
    """

    AXIS_POSITION_PAR_KEY = "ROTATION_AXIS_POSITION"
    """Key used for the axis position in par files"""

    _MANAGED_KEYS = (
        "MODE",
        "POSITION_VALUE",
        "CALC_INPUT_TYPE",
        "USE_SINOGRAM",
        "ANGLE_MODE",
        "SINOGRAM_LINE",
        "AXIS_URL_1",
        "AXIS_URL_2",
        "LOOK_AT_STDMAX",
        "NEAR_WX",
        "FINE_STEP_X",
        "SCALE_IMG2_TO_IMG1",
        "NEAR_POSITION",
        "MOTOR_OFFSET",
        "X_ROTATION_AXIS_PIXEL_POSITION",
        "SINOGRAM_SUBSAMPLING",
        "PADDING_MODE",
        "FLIP_LR",
        "COMPOSITE_OPTS",
        "COR_OPTIONS",
    )

    def __init__(self):
        self.__mode = AxisMode.manual
        """Mode used for defining the COR (center of rotation)"""
        self.__relative_value = None
        """Value of the center of rotation in [0; image_width].
        None is not processing"""
        self.__absolute_value = None
        """Value of the center of rotation in [-image_width/2; image_width/2].
        None is not processing"""
        self.__angle_mode = CorAngleMode.use_0_180
        """Angle to use for radios"""
        self.__x_rotation_axis_pixel_position = None
        """rotation axis position obtained from motor"""
        self.__x_rotation_axis_pos_px_offset = 0.0
        """motor offset to be used to compute the 'estimated_cor' from 'x_rotation_axis_pixel_position'"""
        self.__estimated_cor = 0.0
        """Estimated position of the center of rotation. Given by the user or computed from 'x_rotation_axis_pixel_position' and 'pixel_offset'"""
        self.__axis_url_1 = AxisResource(url=None)
        """first data url to use for axis cor calculation"""
        self.__axis_url_2 = AxisResource(url=None)
        """second data url to use for axis cor calculation"""
        self.__calculation_input_type = AxisCalculationInput.transmission
        """Type of input (emission, absorption, with or without paganin)"""
        self.__sinogram_line = "middle"
        """Line of the radios to use for getting the sinogram"""
        self.__sinogram_subsampling = 10
        """if use sinogram activate, we can use a subsampling to reduce
        computation time"""
        self.__look_at_stdmax = False
        """do the near search at X position which as the max Y column standard
        deviation"""
        self.__composite_window_size = 5
        """do the near search in an X window of size +-near_wx"""
        self.__composite_fine_step = 0.1
        """shift step x for fine shifting image"""
        self.__scale_img2_to_img1 = False
        """do image scaling"""
        self.__padding_mode = None
        self.__frame_width = None
        self.__flip_lr = True
        self.__composite_options = {
            "theta": DEFAULT_CMP_THETA,
            "n_subsampling_y": DEFAULT_CMP_N_SUBSAMPLING_Y,
            "oversampling": DEFAULT_CMP_OVERSAMPLING,
            "take_log": DEFAULT_CMP_TAKE_LOG,
            "near_pos": DEFAULT_CMP_NEAR_POS,
            "near_width": DEFAULT_CMP_NEAR_WIDTH,
        }
        """specific options for composite cor search"""
        self.__extra_cor_options = ""
        """Automatic cor options as str. side, near_pos and near_width are provided independantly"""

    @property
    def mode(self):
        return self.__mode

    @mode.setter
    def mode(self, mode):
        self.__mode = AxisMode.from_value(mode)
        self.changed()

    @property
    def frame_width(self) -> float | None:
        return self.__frame_width

    @frame_width.setter
    def frame_width(self, value):
        if not isinstance(value, (type(None), float, int)):
            raise TypeError(
                f"Value is expected to be None or a float. Not {type(value)}"
            )
        if value is None:
            self.__frame_width = value
        else:
            self.__frame_width = float(value)

    @property
    def angle_mode(self):
        return self.__angle_mode

    @angle_mode.setter
    def angle_mode(self, mode):
        if type(mode) is str:
            mode = CorAngleMode.from_value(mode)

        if self.__angle_mode != mode:
            self.__angle_mode = mode
            self.changed()

    @property
    def relative_cor_value(self):
        return self.__relative_value

    @property
    def absolute_cor_value(self):
        return self.__absolute_value

    def set_relative_value(self, value):
        if not isinstance(value, (int, float, str, type(None))):
            raise TypeError(
                f"value is expected to be an instance of {int}, {float}, {str}, or {None}. {type(value)} provided"
            )
        if value is None:
            changed = self.__relative_value is not None
        elif isinstance(value, str) and value == "...":
            changed = self.__relative_value != "..."
        else:
            changed = self.__relative_value != float(value)
        if changed:
            if value is None or (isinstance(value, str) and value == "..."):
                self.__relative_value = value
                self.__absolute_value = value
            else:
                self.__relative_value = float(value)
                if self.frame_width is not None:
                    self.__absolute_value = relative_pos_to_absolute(
                        relative_pos=self.__relative_value, det_width=self.frame_width
                    )
            self.changed()

    @property
    def estimated_cor(self) -> Side | float | None:
        return self.__estimated_cor

    @estimated_cor.setter
    def estimated_cor(self, value: Side | float | None):
        try:
            value = Side.from_value(value)
        except ValueError:
            pass
        if self.__estimated_cor != value:
            self.__estimated_cor = value
            self.changed()

    @property
    def x_rotation_axis_pos_px_offset(self) -> float:
        return self.__x_rotation_axis_pos_px_offset

    @x_rotation_axis_pos_px_offset.setter
    def x_rotation_axis_pos_px_offset(self, value: float):
        assert isinstance(
            value, float
        ), f"x_rotation_axis_pos_px_offset should be a float. Got {type(value)}"
        self.__x_rotation_axis_pos_px_offset = value

    @property
    def x_rotation_axis_pixel_position(self) -> float | None:
        return self.__x_rotation_axis_pixel_position

    @x_rotation_axis_pixel_position.setter
    def x_rotation_axis_pixel_position(self, value: float | None):
        assert isinstance(
            value, (float, type(None))
        ), f"x_rotation_axis_pixel_position should be None or a float. Got{type(value)}"
        self.__x_rotation_axis_pixel_position = value

    @property
    def axis_url_1(self):
        """the first file to be used for the axis calculation"""
        return self.__axis_url_1

    def _get_rsrc_frm_url(self, old_resource, new_url_):
        """
        Util function to compare new resource / url with an existing one.
        """
        new_resource = None
        changed = False
        if isinstance(new_url_, AxisResource):
            if old_resource != new_url_:
                changed = True
                new_resource = new_url_
        elif isinstance(new_url_, str):
            if new_url_ == "":
                new_resource = AxisResource(None)
            else:
                new_resource = AxisResource(DataUrl(path=new_url_))
            if old_resource != new_resource:
                changed = True
        else:
            assert isinstance(new_url_, DataUrl) or new_url_ is None
            new_resource = AxisResource(new_url_)
            if old_resource != new_resource:
                changed = True
        return new_resource, changed

    @axis_url_1.setter
    def axis_url_1(self, url_):
        new_resource, changed = self._get_rsrc_frm_url(
            old_resource=self.__axis_url_1, new_url_=url_
        )
        if changed is True:
            self.__axis_url_1 = new_resource
            self.axis_urls_changed()

    @property
    def axis_url_2(self):
        """the second file to be used for the axis calculation"""
        return self.__axis_url_2

    @axis_url_2.setter
    def axis_url_2(self, url_):
        new_resource, changed = self._get_rsrc_frm_url(
            old_resource=self.__axis_url_2, new_url_=url_
        )
        if changed is True:
            self.__axis_url_2 = new_resource
            self.axis_urls_changed()

    @property
    def flip_lr(self):
        return self.__flip_lr

    @flip_lr.setter
    def flip_lr(self, flip: bool):
        if not isinstance(flip, bool):
            raise TypeError("flip should be a boolean")
        self.__flip_lr = flip

    @property
    def projection_type(self):
        return self.__calculation_input_type.value.projection_type

    @property
    def paganin_preproc(self):
        return self.__calculation_input_type.value.paganin

    @property
    def calculation_input_type(self):
        return self.__calculation_input_type

    @calculation_input_type.setter
    def calculation_input_type(self, type_):
        assert isinstance(type_, AxisCalculationInput)
        value = AxisCalculationInput.from_value(type_)
        if value != self.__calculation_input_type:
            self.__calculation_input_type = type_
            self.changed()

    @property
    def sinogram_line(self):
        return self.__sinogram_line

    @sinogram_line.setter
    def sinogram_line(self, line):
        if line == "":
            if self.__sinogram_line is not None:
                self.__sinogram_line = None
                self.changed()
        else:
            if line == "middle":
                line = line
            elif line is not None:
                line = int(line)
            if self.__sinogram_line != line:
                self.__sinogram_line = line
                self.changed()

    @property
    def sinogram_subsampling(self) -> int:
        """Subsample radio to speed up processing of the sinogram generation"""
        return self.__sinogram_subsampling

    @sinogram_subsampling.setter
    def sinogram_subsampling(self, subsampling: int) -> None:
        subsampling = int(subsampling)
        if subsampling != self.__sinogram_subsampling:
            self.__sinogram_subsampling = subsampling
            self.changed()

    @property
    def look_at_stdmax(self):
        return self.__look_at_stdmax

    @look_at_stdmax.setter
    def look_at_stdmax(self, stdmax):
        self.__look_at_stdmax = stdmax

    @property
    def composite_window_size(self):
        return self.__composite_window_size

    @composite_window_size.setter
    def composite_window_size(self, width):
        if self.__composite_window_size != width:
            self.__composite_window_size = width
            self.changed()

    @property
    def composite_fine_step(self):
        """Fine step along x axis"""
        return self.__composite_fine_step

    @composite_fine_step.setter
    def composite_fine_step(self, step_size):
        if self.__composite_fine_step != step_size:
            self.__composite_fine_step = step_size
            self.changed()

    @property
    def scale_img2_to_img1(self):
        return self.__scale_img2_to_img1

    @scale_img2_to_img1.setter
    def scale_img2_to_img1(self, scale):
        if self.__scale_img2_to_img1 != scale:
            self.__scale_img2_to_img1 = scale
            self.changed()

    @property
    def extra_cor_options(self) -> str:
        return self.__extra_cor_options

    @extra_cor_options.setter
    def extra_cor_options(self, cor_options: str):
        if not isinstance(cor_options, str):
            raise TypeError(f"{type(cor_options)} provided when {str} expected")
        else:
            clean_opts_str = (
                cor_options.replace(" ", "").strip(";").rstrip(";").replace(";", " ; ")
            )
            self.__extra_cor_options = clean_opts_str
            self.changed()

    @property
    def composite_options(self) -> dict:
        """return specific options for composite cor search"""
        return self.__composite_options

    @composite_options.setter
    def composite_options(self, opts: dict) -> None:
        """
        :param opts: options to use for the composite COR search
        :raises:
            * KeyError if some provided keys are not handled
            * TypeError if opts is not an instance of dictionary
        """
        if not isinstance(opts, dict):
            raise TypeError(
                f"opts is expected to be an instance of dict not {type(opts)}"
            )
        # insure backward compatibility
        if "subsampling_y" in opts:
            opts["n_subsampling_y"] = opts.pop("subsampling_y")
        for key in opts.keys():
            if key not in (
                "theta",
                "oversampling",
                "n_subsampling_y",
                "take_log",
                "near_pos",
                "near_width",
            ):
                raise KeyError(f"{key} is not recognized")
        self.__composite_options = opts

    @property
    def padding_mode(self):
        return self.__padding_mode

    @padding_mode.setter
    def padding_mode(self, mode):
        if self.__padding_mode != mode:
            self.__padding_mode = mode
            self.changed()

    def changed(self):
        """callback to overwrite when the parameter value changed"""
        pass

    def n_url(self):
        """

        :return: number of available url from url_1, url_2
        """
        n_url = 0
        if self.axis_url_1 and self.axis_url_1.url:
            n_url += 1
        if self.axis_url_2 and self.axis_url_2.url:
            n_url += 1
        return n_url

    def to_dict(self):
        # keep octave compatibility
        axis_urls_1 = self.axis_url_1.url
        if axis_urls_1 is None:
            axis_urls_1 = ""
        else:
            axis_urls_1 = axis_urls_1.path()
        axis_urls_2 = self.axis_url_2.url
        if axis_urls_2 is None:
            axis_urls_2 = ""
        else:
            axis_urls_2 = axis_urls_2.path()

        _dict = {
            "MODE": self.mode.value,
            "POSITION_VALUE": self.relative_cor_value,
            "CALC_INPUT_TYPE": self.calculation_input_type.to_dict(),
            "ANGLE_MODE": self.angle_mode.value,
            "SINOGRAM_LINE": (
                self.sinogram_line if self.mode.requires_sinogram_index() else ""
            ),
            "SINOGRAM_SUBSAMPLING": self.sinogram_subsampling,
            "AXIS_URL_1": axis_urls_1,
            "AXIS_URL_2": axis_urls_2,
            "LOOK_AT_STDMAX": self.look_at_stdmax,
            "NEAR_WX": self.composite_window_size,
            "FINE_STEP_X": self.composite_fine_step,
            "SCALE_IMG2_TO_IMG1": self.scale_img2_to_img1,
            "NEAR_POSITION": (
                self.estimated_cor.value
                if isinstance(self.estimated_cor, Side)
                else self.estimated_cor
            ),
            "PADDING_MODE": self.padding_mode,
            "FLIP_LR": self.flip_lr,
            "COMPOSITE_OPTS": self.composite_options,
            "COR_OPTIONS": self.extra_cor_options,
            "MOTOR_OFFSET": self.x_rotation_axis_pos_px_offset,
            "X_ROTATION_AXIS_PIXEL_POSITION": self.x_rotation_axis_pixel_position,
        }
        return _dict

    @staticmethod
    def from_dict(_dict):
        axis = AxisRP()
        axis.load_from_dict(_dict=_dict)
        return axis

    def load_from_dict(self, _dict):
        # Convert managed keys to upper case
        _dict = {
            key.upper() if key.upper() in self._MANAGED_KEYS else key: value
            for key, value in _dict.items()
        }

        if "MODE" in _dict:
            self.mode = _dict["MODE"]
        if "POSITION_VALUE" in _dict:
            self.set_relative_value(_dict["POSITION_VALUE"])
        if "CALC_INPUT_TYPE" in _dict:
            self.calculation_input_type = AxisCalculationInput.from_value(
                _dict["CALC_INPUT_TYPE"]
            )
        if "ANGLE_MODE" in _dict:
            self.angle_mode = CorAngleMode.from_value(_dict["ANGLE_MODE"])
        if "SINOGRAM_LINE" in _dict:
            self.sinogram_line = _dict["SINOGRAM_LINE"]
        if "AXIS_URL_1" in _dict:
            self.axis_url_1 = _dict["AXIS_URL_1"]
        if "AXIS_URL_2" in _dict:
            self.axis_url_2 = _dict["AXIS_URL_2"]
        if "LOOK_AT_STDMAX" in _dict:
            self.look_at_stdmax = _dict["LOOK_AT_STDMAX"]
        if "NEAR_WX" in _dict:
            self.composite_window_size = _dict["NEAR_WX"]
        if "FINE_STEP_X" in _dict:
            self.composite_fine_step = _dict["FINE_STEP_X"]
        if "SCALE_IMG2_TO_IMG1" in _dict:
            self.scale_img2_to_img1 = _dict["SCALE_IMG2_TO_IMG1"]
        if "NEAR_POSITION" in _dict:
            self.estimated_cor = _dict["NEAR_POSITION"]
        if "SINOGRAM_SUBSAMPLING" in _dict:
            self.sinogram_subsampling = _dict["SINOGRAM_SUBSAMPLING"]
        if "PADDING_MODE" in _dict:
            self.padding_mode = _dict["PADDING_MODE"]
        if "FLIP_LR" in _dict:
            self.flip_lr = bool(_dict["FLIP_LR"])
        if "COMPOSITE_OPTS" in _dict:
            self.composite_options = _dict["COMPOSITE_OPTS"]
        self.extra_cor_options = _dict.get("COR_OPTIONS", "")

    def copy(self, axis_params, copy_axis_url=True, copy_flip_lr=True):
        assert isinstance(axis_params, AxisRP)
        self.mode = axis_params.mode
        self.frame_width = axis_params.frame_width
        self.set_relative_value(axis_params.relative_cor_value)
        self.calculation_input_type = axis_params.calculation_input_type
        self.angle_mode = axis_params.angle_mode
        self.sinogram_line = axis_params.sinogram_line
        self.sinogram_subsampling = axis_params.sinogram_subsampling
        self.look_at_stdmax = axis_params.look_at_stdmax
        self.composite_window_size = axis_params.composite_window_size
        self.composite_fine_step = axis_params.composite_fine_step
        self.scale_img2_to_img1 = axis_params.scale_img2_to_img1
        self.estimated_cor = axis_params.estimated_cor
        self.x_rotation_axis_pixel_position = axis_params.x_rotation_axis_pixel_position
        self.x_rotation_axis_pos_px_offset = axis_params.x_rotation_axis_pos_px_offset
        self.padding_mode = axis_params.padding_mode
        self.composite_options = axis_params.composite_options
        self.extra_cor_options = axis_params.extra_cor_options
        if copy_axis_url:
            self.axis_url_1 = axis_params.axis_url_1
            self.axis_url_2 = axis_params.axis_url_2
        if copy_flip_lr:
            self.flip_lr = axis_params.flip_lr

    def __str__(self):
        return str(self.to_dict())

    def axis_urls_changed(self):
        """Callback when the axis url change"""
        pass

    def get_simple_str(self):
        """
        special information as a str for mode able to handle both sinogram and radios
        """
        results = f"{self.mode.value}"
        if self.mode in (
            AxisMode.growing_window_radios,
            AxisMode.growing_window_sinogram,
            AxisMode.sliding_window_radios,
            AxisMode.sliding_window_sinogram,
        ):
            extra_info = f"side: {self.estimated_cor}"
            results = ", ".join((results, extra_info))
        return results

    def get_nabu_cor_options_as_dict(self) -> str:
        options = {}
        # provide side. Scalar if a first guess is provided else a value in ('left', 'right', 'center', 'all')
        if (
            isinstance(self.estimated_cor, Side)
            and AXIS_MODE_METADATAS[self.mode].valid_sides
        ):
            options["side"] = self.estimated_cor.value
        elif AXIS_MODE_METADATAS[self.mode].allows_estimated_cor_as_numerical_value:
            options["side"] = self.estimated_cor

        if self.mode in (AxisMode.composite_coarse_to_fine, AxisMode.near):
            options["near_width"] = self.composite_options.get(
                "near_width", DEFAULT_CMP_NEAR_WIDTH
            )
            options["theta_interval"] = self.composite_options.get(
                "theta_interval", DEFAULT_CMP_THETA
            )
            options["oversampling"] = self.composite_options.get(
                "oversampling", DEFAULT_CMP_OVERSAMPLING
            )
            options["n_subsampling_y"] = self.composite_options.get(
                "n_subsampling_y", DEFAULT_CMP_N_SUBSAMPLING_Y
            )
            options["take_log"] = self.composite_options.get(
                "take_log", DEFAULT_CMP_TAKE_LOG
            )

        # provide radio indices or sinogram index
        if self.mode.requires_radio_indices:
            # warning: nabu expect radio angles to be in rad, in nxtomo and tomwer they are in degrees
            options["radio_angles"] = (
                numpy.deg2rad(self.axis_url_1.angle or 0.0),
                numpy.deg2rad(self.axis_url_2.angle or 180.0),
            )
        if self.mode.requires_sinogram_index:
            options["slice_idx"] = self.sinogram_line

        # append "extra_cor_options" to already handled cor options
        # expected values: low_pass, high_pass
        extra_cor_options = self.extra_cor_options.replace(" ", "")

        if extra_cor_options != "":
            for opt in self.extra_cor_options.replace(" ", "").split(";"):
                if len(opt.split("=")) == 2:
                    key, value = opt.split("=")
                    if key in ("low_pass", "high_pass"):
                        value = float(value)
                        options[key] = value
                    else:
                        _logger.warning(
                            f"key {key} is not recognized by nabu. Ignore it."
                        )
                else:
                    _logger.info(f"ignore option {opt}. Invalid syntax")
        return options



import pytest

from nabu.stitching.config import StitchingType
from nabu.stitching.tests import test_y_preprocessing_stitching
from nabu.stitching.tests import test_z_preprocessing_stitching
from nabu.stitching.tests import test_z_postprocessing_stitching

from tomwer.app.stitching.common import MainWindow
from tomwer.core.scan.nxtomoscan import NXtomoScan
from tomwer.core.volume.volumefactory import VolumeFactory
from tomoscan.esrf.volume import HDF5Volume as _tomoscan_HDF5Volume
from tomwer.tests.conftest import qtapp  # noqa F401


def _create_objects_for_stitching(stitching_type: StitchingType, output_dir) -> tuple:
    """
    create tomo object (scans or volumes) and return them with there position along the axis to stitch
    """

    def cast_tomoscan_scans_to_tomwer_scans(scans: tuple) -> tuple:
        return tuple(
            [NXtomoScan(scan=scan.master_file, entry=scan.entry) for scan in scans]
        )

    def cast_tomoscan_volumes_to_tomwer_volumes(volumes: tuple) -> tuple:
        return tuple(
            [
                VolumeFactory.create_tomo_object_from_identifier(
                    volume.get_identifier().to_str()
                )
                for volume in volumes
            ]
        )

    stitching_type = StitchingType.from_value(stitching_type)
    if stitching_type is StitchingType.Y_PREPROC:
        nxtomos, positions, _ = test_y_preprocessing_stitching.build_nxtomos(
            output_dir=output_dir, flip_lr=False, flip_ud=False
        )
        return cast_tomoscan_scans_to_tomwer_scans(nxtomos), positions
    elif stitching_type is StitchingType.Z_PREPROC:
        nxtomos, positions, _ = test_z_preprocessing_stitching.build_nxtomos(
            output_dir=output_dir
        )
        return cast_tomoscan_scans_to_tomwer_scans(nxtomos), positions
    elif stitching_type is StitchingType.Z_POSTPROC:
        volumes, positions, _ = test_z_postprocessing_stitching.build_volumes(
            output_dir=output_dir, volume_class=_tomoscan_HDF5Volume
        )
        return cast_tomoscan_volumes_to_tomwer_volumes(volumes), positions
    else:
        raise NotImplementedError()


@pytest.mark.parametrize("stitching_type", StitchingType)
def test_stitching_on_projections(stitching_type, tmp_path, qtapp):  # noqa F401
    if stitching_type in (StitchingType.Z_PREPROC, StitchingType.Z_POSTPROC):
        axis = 0
    elif stitching_type is StitchingType.Y_PREPROC:
        axis = 1
    else:
        raise NotImplementedError

    widget = MainWindow(axis=axis)
    output_dir = tmp_path / "input_objs"
    output_dir.mkdir()
    tomo_objects, positions = _create_objects_for_stitching(stitching_type, output_dir)
    positions_dict = {}
    for tomo_object, position in zip(tomo_objects, positions):
        widget.addTomoObj(tomo_object)
        positions_dict[tomo_object.get_identifier().to_str()] = position

    pos_widget = (
        widget._mainWindow._stitchingConfigWindow._editTomoObjFirstAxisPositionsWidget._tomoObjtoTomoObjPosWidget
    )
    # make sure the position over the axis to stitch are valid
    for tomo_obj, w in pos_widget.items():
        assert positions_dict[tomo_obj] == w.getOriginalValue()

    stitching_configuration = widget.getStitchingConfiguration()
    second_axis = widget._mainWindow._stitchingConfigWindow.second_axis

    # only the axis of the stitching is defined. When launching the widget we don't know if this is pre-processing or post-processing
    assert stitching_configuration["stitching"]["type"][0] == stitching_type.value[0]
    assert len(stitching_configuration["inputs"]["input_datasets"]) == len(tomo_objects)
    assert stitching_configuration["stitching"][f"axis_{axis}_pos_px"] == sorted(
        positions, reverse=True
    )
    # at the moment the 'overlap' size is only handled and provided for the 'main axis (stitching axis)
    assert "overlap_size" in stitching_configuration["stitching"][f"axis_{axis}_params"]
    assert (
        "overlap_size"
        not in stitching_configuration["stitching"][f"axis_{second_axis}_params"]
    )
    assert "axis_2_params" not in stitching_configuration["stitching"]



from __future__ import annotations

import time

from orangewidget import gui

# from orangewidget.widget import Input, Output, OWBaseWidget
from orangecontrib.tomwer.orange.managedprocess import TomwerWithStackStack
from ewokscore.missing_data import MISSING_DATA
from ewoksorange.gui.orange_imports import Input

from processview.core.manager import DatasetState, ProcessManager
from processview.core.superviseprocess import SuperviseProcess
from silx.gui import qt
from silx.gui.utils.concurrent import submitToQtMainThread

from tomwer.core.futureobject import FutureTomwerObject
from tomwer.gui.cluster.supervisor import (
    FutureTomwerScanObserverWidget as _FutureTomwerScanObserverWidget,
)
from tomwer.core.process.cluster.supervisor import FutureSupervisorTask


class FutureSupervisorOW(
    TomwerWithStackStack,
    ewokstaskclass=FutureSupervisorTask,
):
    """
    Orange widget to define a slurm cluster as input of other
    widgets (based on nabu for now)
    """

    name = "future supervisor"
    id = "orange.widgets.tomwer.cluster.FutureSupervisorOW.FutureSupervisorOW"
    description = "Observe slurm job registered."
    icon = "icons/slurmobserver.svg"
    priority = 22
    keywords = [
        "tomography",
        "tomwer",
        "slurm",
        "observer",
        "cluster",
        "job",
        "sbatch",
        "supervisor",
        "future",
    ]

    want_main_area = True
    want_control_area = False
    resizing_enabled = True

    class Inputs:
        # redefine the input to allow multiple and default
        future_tomo_obj = Input(
            name="future_tomo_obj",
            type=FutureTomwerObject,
            doc="data with some remote processing",
            multiple=True,
            default=True,
        )

    def __init__(self, parent=None):
        super().__init__(parent)
        # gui
        layout = gui.vBox(self.mainArea, self.name).layout()
        self._widget = FutureTomwerObjectObserverWidget(
            parent=self, name=self.windowTitle()
        )
        layout.addWidget(self._widget)

        # connect signal / slot
        self._widget.observationTable.model().sigStatusUpdated.connect(
            self._convertBackAutomatically
        )
        self._widget.sigConversionRequested.connect(self._convertBack)

    def convertWhenFinished(self):
        return self._widget.convertWhenFinished()

    def _convertBackAutomatically(self, future_tomo_obj, status):
        if not isinstance(future_tomo_obj, FutureTomwerObject):
            raise TypeError(
                f"future_tomo_obj is expected to be an instance of {FutureTomwerObject} and not {type(future_tomo_obj)}"
            )
        if status in ("finished", "completed") and self.convertWhenFinished():
            self._convertBack(future_tomo_obj)

    def _convertBack(self, future_tomo_obj):
        if not isinstance(future_tomo_obj, FutureTomwerObject):
            raise TypeError(
                f"future_tomo_obj is expected to be an instance of {FutureTomwerObject} and not {type(future_tomo_obj)}"
            )
        self._widget.removeFutureTomoObj(future_tomo_obj=future_tomo_obj)
        self.execute_ewoks_task()

    def handleNewSignals(self) -> None:
        """Invoked by the workflow signal propagation manager after all
        signals handlers have been called.

        note: this widget can receive two signals: 'dataset' and 'colormap'. The 'colormap' is handled by
              orange directly while the 'dataset' signal is handled by the ewoks task.
              This function will be only triggered when the 'dataset' signal is send
        """
        # update GUI from received future_tomo_obj
        # warning: this code will work because the task has only one input.
        # so we can pass it directly to the widget.
        # this won't be the case the task can have several input.
        future_tomo_obj = self.get_task_input_value("future_tomo_obj", MISSING_DATA)
        if future_tomo_obj is not MISSING_DATA:
            self._widget.addFutureTomoObj(future_tomo_obj=future_tomo_obj)

    @Inputs.future_tomo_obj
    def add(self, future_tomo_obj, signal_id=None):
        # required because today ewoksorange is not handling multiple inputs
        self.set_dynamic_input("future_tomo_obj", future_tomo_obj)


class FutureTomwerObjectObserverWidget(
    _FutureTomwerScanObserverWidget, SuperviseProcess
):
    """add dataset state notification (ProcessManager) to the original FutureTomwerScanObserverWidget"""

    REFRESH_FREQUENCE = 10
    """time between call to updateView"""

    def __init__(self, name, parent=None):
        super().__init__(parent=parent)
        self.name = name
        self._updateThread = _RefreshThread(
            callback=self.updateView, refresh_frequence=self.REFRESH_FREQUENCE
        )
        self.destroyed.connect(self.stopRefresh)
        self._updateThread.start()

    def stopRefresh(self):
        if self._updateThread is not None:
            self._updateThread.stop()
            self._updateThread.wait(self.REFRESH_FREQUENCE + 1)
            self._updateThread = None

    def close(self):
        self.stopRefresh()
        super().close()

    def addFutureTomoObj(self, future_tomo_obj: FutureTomwerObject):
        super().addFutureTomoObj(future_tomo_obj)
        self._updateTomoObjSupervisor(future_tomo_obj)

    def removeFutureTomoObj(self, future_tomo_obj: FutureTomwerObject):
        self._updateTomoObjSupervisor(future_tomo_obj)
        super().removeFutureTomoObj(future_tomo_obj)

    def _updateTomoObjSupervisor(self, future_tomo_obj):
        r_id = future_tomo_obj.process_requester_id
        if r_id is not None:
            requester_name = ProcessManager().get_process(r_id).name
        else:
            requester_name = "unknow"
        details = f"job spawn by {requester_name}"
        if future_tomo_obj is None:
            return
        elif future_tomo_obj.status == "error":
            state = DatasetState.FAILED
        elif future_tomo_obj.status == "pending":
            details = "\n".join([details, "pending"])
            state = DatasetState.PENDING
        elif future_tomo_obj.status in ("finished", "completed"):
            details = future_tomo_obj.logs or "no log found"
            state = DatasetState.SUCCEED
        elif future_tomo_obj.status == "running":
            details = "\n".join([details, "running"])
            state = DatasetState.ON_GOING
        elif future_tomo_obj.status == "cancelled":
            details = "\n".join([details, "job cancelled"])
            state = DatasetState.SKIPPED
        elif future_tomo_obj.status is None:
            return
        else:
            raise ValueError(
                f"future scan status '{future_tomo_obj.status}' is not managed, {type(future_tomo_obj.status)}"
            )
        ProcessManager().notify_dataset_state(
            dataset=future_tomo_obj.tomo_obj,
            process=self,
            state=state,
            details=details,
        )

    def _updateStatus(self, future_tomo_obj):
        self._updateTomoObjSupervisor(future_tomo_obj)
        super()._updateStatus(future_tomo_obj)


class _RefreshThread(qt.QThread):
    """Simple thread to call a refresh callback each refresh_frequence (seconds)"""

    TIME_BETWEEN_LOOP = 1.0

    def __init__(self, callback, refresh_frequence) -> None:
        super().__init__()
        self._callback = callback
        self._refresh_frequence = refresh_frequence
        self._stop = False

    def stop(self):
        self._stop = True
        self._callback = None

    def run(self):
        w_t = self._refresh_frequence + self.TIME_BETWEEN_LOOP

        while not self._stop:
            if w_t <= 0:
                if self._callback is not None:
                    try:
                        submitToQtMainThread(self._callback)
                    except AttributeError:
                        # can happen when closing
                        pass
                w_t = self._refresh_frequence + self.TIME_BETWEEN_LOOP
            w_t -= self.TIME_BETWEEN_LOOP
            time.sleep(self.TIME_BETWEEN_LOOP)



"""some utils to move a path to process_data or to raw_data"""

import os
from nxtomomill.converter.hdf5.utils import (
    RAW_DATA_DIR_NAME,
    PROCESSED_DATA_DIR_NAME,
)


def to_raw_data_path(file_path: str):
    """convert a path to 'RAW_DATA' when possible"""
    split_path = file_path.split(os.sep)
    # reverse it to find the lower level value of 'PROCESSED_DATA_DIR_NAME' if by any 'chance' has several in the path
    # in this case this is most likely what we want
    split_path = split_path[::-1]
    # check if already contain in a "RAW_DATA_DIR_NAME" directory
    try:
        index_raw_data = split_path.index(RAW_DATA_DIR_NAME)
    except ValueError:
        index_raw_data = None

    try:
        index_processed_data = split_path.index(PROCESSED_DATA_DIR_NAME)
    except ValueError:
        # if the value is not in the list
        pass
    else:
        if index_raw_data is None or index_raw_data > index_processed_data:
            # make sure we are not already in a 'PROCESSED_DATA' directory. Not sure it will never happen but safer
            split_path[index_processed_data] = RAW_DATA_DIR_NAME

    # reorder path to original
    split_path = split_path[::-1]
    return os.sep.join(split_path)


def to_processed_data_path(file_path: str):
    """convert a path to 'PROCESSED_DATA' when possible"""
    split_path = file_path.split(os.sep)
    # reverse it to find the lower level value of '_RAW_DATA_DIR_NAME' if by any 'chance' has several in the path
    # in this case this is most likely what we want
    split_path = split_path[::-1]
    # check if already contain in a "PROCESSED_DATA" directory
    try:
        index_processed_data = split_path.index(PROCESSED_DATA_DIR_NAME)
    except ValueError:
        index_processed_data = None

    try:
        index_raw_data = split_path.index(RAW_DATA_DIR_NAME)
    except ValueError:
        # if the value is not in the list
        pass
    else:
        if index_processed_data is None or index_raw_data < index_processed_data:
            # make sure we are not already in a 'PROCESSED_DATA' directory. Not sure it will never happen but safer
            split_path[index_raw_data] = PROCESSED_DATA_DIR_NAME

    # reorder path to original
    split_path = split_path[::-1]
    return os.sep.join(split_path)


def file_is_on_processed_data(file_path: str):
    """Util that raises an error if path is not in PROCESSED_DATA folder"""
    split_path = file_path.split(os.sep)
    # reverse it to find the lower level value of 'PROCESSED_DATA_DIR_NAME' if by any 'chance' has several in the path
    # in this case this is most likely what we want
    split_path = split_path[::-1]
    # check if already contain in a "RAW_DATA_DIR_NAME" directory
    try:
        index_raw_data = split_path.index(RAW_DATA_DIR_NAME)
    except ValueError:
        index_raw_data = None

    try:
        index_processed_data = split_path.index(PROCESSED_DATA_DIR_NAME)
    except ValueError:
        # if the value is not in the list
        index_processed_data = None
    else:
        if index_raw_data is not None and index_raw_data < index_processed_data:
            return False
    return index_processed_data is not None



from __future__ import annotations

from copy import copy
import os
import numpy
import logging

from silx.gui import qt
from silx.gui.plot import Plot2D
from silx.io.dictdump import h5todict
from silx.io.url import DataUrl
from silx.io.utils import open as open_hdf5
from silx.gui.dialog.DataFileDialog import DataFileDialog

from tomoscan.esrf.scan.utils import cwd_context
from tomoscan.framereducer.target import REDUCER_TARGET

from tomwer.io.utils import get_default_directory

_logger = logging.getLogger(__name__)


class ReduceDarkFlatSelectorTableWidget(qt.QWidget):
    """
    Table widget used to register and select a list of reduces frames
    """

    sigActiveChanged = qt.Signal(object)
    """Signal emit when the active frame changed (the one to be displayed)"""

    sigUpdated = qt.Signal()
    """Signal emit when the table is updated"""

    def __init__(self, parent: qt.QWidget | None = None) -> None:
        super().__init__(parent)
        self._iDict = 0
        self._dicts = {}
        # key are 'dict name', value are dict of the reduced frame (acquisition index as key, 2D numpy array as value).
        self.setLayout(qt.QVBoxLayout())
        self._tree = qt.QTreeWidget(self)
        self._tree.setHeaderLabels(("reduce frames", "acq index", "select"))
        self._tree.header().setStretchLastSection(False)
        self._tree.header().setSectionResizeMode(0, qt.QHeaderView.Stretch)
        self._tree.setColumnWidth(1, 80)
        self._tree.setColumnWidth(2, 50)
        self.layout().addWidget(self._tree)

        self.setAcceptDrops(True)

        # connect signal / slot
        self._tree.currentItemChanged.connect(self._updateActiveChanged)

    def _updateActiveChanged(self, item, column: int):
        frame = item.data(0, qt.Qt.UserRole)
        self.sigActiveChanged.emit(frame)

    def popDictIndex(self) -> int:
        self._iDict = self._iDict + 1
        return self._iDict - 1

    def addReduceFrames(self, reduced_frames: dict, selected: tuple = ()):
        """
        Add the reduced frames in the interface.

        :param reduced_frames: frame index as key and numpy.ndarray as value
        :param selected: if the frame index is in selected tuple then this item will be automatically selected
        """
        reduced_frames = copy(reduced_frames)
        dict_key = (
            reduced_frames.pop("reduce_frames_name", None)
            or f"reduced frames #{self.popDictIndex()}"
        )
        # if the key already exists: remove it (avoid )
        if dict_key in self._dicts:
            self._removeReduceFramesByLabel(dict_key)
        self._dicts[dict_key] = reduced_frames

        topLevelItem = qt.QTreeWidgetItem(self._tree)
        topLevelItem.setText(0, dict_key)

        for frame_index, frame in reduced_frames.items():
            if not isinstance(frame, numpy.ndarray):
                _logger.error(
                    f"frame are expected to be numpy.ndarray. Get {type(frame)} instead - will not add it"
                )
                continue
            if frame.ndim != 2:
                _logger.error(f"frame are expected to be 2D. Get {frame.shape} instead")
                continue

            # frame info
            frameInfo = qt.QTreeWidgetItem(topLevelItem)
            frameInfo.setText(0, f"shape: {frame.shape}, dtype {frame.dtype}")
            frameInfo.setData(0, qt.Qt.UserRole, frame)
            # frame index
            indexItem = qt.QLineEdit(self._tree)
            # index can be provided as absolute indexes (integers)
            validator = qt.QRegExpValidator(
                qt.QRegExp(r"([+-]?([0-9]*[.])?[0-9]+)*[r]"), self
            )
            indexItem.setValidator(validator)
            indexItem.setText(str(frame_index))
            self._tree.setItemWidget(frameInfo, 1, indexItem)
            indexItem.setToolTip(
                "Index of the reduced frames in the acquisition. Value can be provided as an absolute value - integer or as a relative value. In this case we expect users to provide a value between [0 and 1.0[ and postfix by 'r'"
            )
            # item selection
            selectCB = qt.QCheckBox(self._tree)
            selectCB.setChecked(frame_index in selected)
            self._tree.setItemWidget(frameInfo, 2, selectCB)
        self.sigUpdated.emit()

    def clear(self):
        self._dicts.clear()
        self._tree.clear()
        self.sigUpdated.emit()

    def getConfiguration(self) -> tuple:
        """
        return configuration as a tuple of 'reduced frames'. Each reduced_frames are provided with the following:
        * 'reduce_frames_name': str: name of the reduced frame
        * 'reduce_frames': tuple of dict for each reduce frame. This second dict contains:

            * index: index of this frame in the acquisition
            * data: the reduce frame as a 2D array
            * selected: is the frame selected by the user

        """
        config = []
        for i_child in range(self._tree.topLevelItemCount()):
            child = self._tree.topLevelItem(i_child)
            reduce_frames_infos = []
            for j_child in range(child.childCount()):
                sub_child = child.child(j_child)
                reduce_frames_infos.append(
                    {
                        "data": sub_child.data(0, qt.Qt.UserRole),
                        "index": self._get_reduce_frame_index(
                            self._tree.itemWidget(sub_child, 1).text()
                        ),
                        "selected": self._tree.itemWidget(sub_child, 2).isChecked(),
                    }
                )
            config.append(
                {
                    "reduce_frames_name": child.text(0),
                    "reduce_frames": reduce_frames_infos,
                }
            )
        return tuple(config)

    def setConfiguration(self, config: tuple | None):
        if config is None:
            return
        self.clear()
        for reduce_frames in config:
            frames_dict = {
                "reduce_frames_name": reduce_frames.get("reduce_frames_name", None)
            }
            selected = []
            for reduce_frame in reduce_frames["reduce_frames"]:
                assert isinstance(
                    reduce_frame, dict
                ), f"reduce_frames is expected to contain a list of dict. Get {type(reduce_frame)} instead"
                frame_index = reduce_frame["index"]
                assert isinstance(frame_index, int) or (
                    isinstance(frame_index, str) and frame_index.endswith("r")
                ), f"frame index should be an int or a str ending by 'r', get {type(frame_index)}"
                frames_dict[frame_index] = reduce_frame["data"]
                if reduce_frame["selected"]:
                    selected.append(reduce_frame["index"])
            self.addReduceFrames(frames_dict, selected)
        self.sigUpdated.emit()

    @staticmethod
    def _get_reduce_frame_index(index_as_str):
        """
        simple util to return expected frame index depending if the index is provided as relative or as absolute.
        """
        if index_as_str.endswith("r"):
            return index_as_str
        else:
            return int(index_as_str)

    def getSelectedReduceFrames(self) -> dict:
        """
        Return the selected reduced frame as dict
        """
        reduce_frames = {}
        for i_child in range(self._tree.topLevelItemCount()):
            child = self._tree.topLevelItem(i_child)
            for j_child in range(child.childCount()):
                sub_child = child.child(j_child)
                is_selected = self._tree.itemWidget(sub_child, 2).isChecked()
                if not is_selected:
                    continue

                frame_index = self._get_reduce_frame_index(
                    self._tree.itemWidget(sub_child, 1).text()
                )
                data = sub_child.data(0, qt.Qt.UserRole)
                assert isinstance(
                    data, numpy.ndarray
                ), f"frame are expected to be numpy.ndarray. Get {type(data)} instead"
                assert (
                    data.ndim == 2
                ), f"frame are expected to be 2D. Get {data.shape} instead"
                if frame_index in reduce_frames:
                    _logger.error(
                        f"frame index {frame_index} defined twice. Ignore last one"
                    )
                reduce_frames[frame_index] = data
        return reduce_frames

    def clearSelection(self):
        """
        Uncheck any check Combobox found
        """
        for i_child in range(self._tree.topLevelItemCount()):
            child = self._tree.topLevelItem(i_child)
            for j_child in range(child.childCount()):
                sub_child = child.child(j_child)
                if self._tree.itemWidget(sub_child, 2).isChecked():
                    self._tree.itemWidget(sub_child, 2).setChecked(False)

    def _removeSelected(self):
        # note: simplest way to remove item it to reset the configuration
        configuration = self.getConfiguration()
        configuration = filter_unselected_reduced_frames(configuration=configuration)
        self.setConfiguration(configuration)
        self.sigUpdated.emit()

    def _removeReduceFramesByLabel(self, label: str):
        """
        remove the label named 'label' if exists.
        The simplest way to remove reduced frames are by reseting the configuration for now
        """
        configuration = list(self.getConfiguration())
        new_configuration = tuple(
            filter(
                lambda my_dict: my_dict.get("reduce_frames_name", None) != label,
                configuration,
            ),
        )

        if len(configuration) != len(new_configuration):
            self.setConfiguration(new_configuration)

    def _guessReduceFramesFromFile(self, file_path: str) -> tuple:
        if not os.path.exists(file_path):
            _logger.error(f"file doesn't exists ({file_path})")

        with open_hdf5(file_path) as h5f:
            entries = tuple(h5f.keys())

        res = []
        for entry in entries:
            res.extend(
                self.get_reduce_frames(
                    DataUrl(
                        file_path=file_path,
                        data_path=entry,
                    )
                )
            )
        return tuple(res)

    @staticmethod
    def get_reduce_frames(url: DataUrl) -> tuple:
        """
        try to guess location of darks / flats according to provided url.
        The idea is to be more robust if the user provide a file or an higher level data path

        :warning: Return a tuple and not a dict. Because in the case the user provide an entry which contains
                  a 'darks' and a flats' groups we want to return both reduced_frames.
                  And we cannot return them as a dict as they can have the same index - used as dict key
        """
        if not isinstance(url, DataUrl):
            raise TypeError(f"url is expected to be a DataUrl. Get {type(url)} instead")
        if not os.path.exists(url.file_path()):
            _logger.error(f"file doesn't exists ({url.file_path()})")
            return tuple()

        result = []
        with cwd_context(url.file_path()):
            reduced_info_dict = h5todict(
                h5file=url.file_path(),
                path=url.data_path(),
            )
            for target in REDUCER_TARGET.values():
                if target in reduced_info_dict:
                    reduced_frames = reduced_info_dict[target]
                    reduced_frames["reduce_frames_name"] = (
                        f"{target} from {url.data_path()}@{os.path.basename(url.file_path())}"
                    )
                    result.append(reduced_frames)

            if len(reduced_frames) == 0:
                # else we consider the data_path is the valid one
                reduced_frames["reduce_frames_name"] = (
                    f"{url.data_path()}@{os.path.basename(url.file_path())}"
                )
                result.append(reduced_frames)

        return tuple(result)

    # drag / drop handling

    def dropEvent(self, event):
        if event.mimeData().hasFormat("text/uri-list"):
            for url in event.mimeData().urls():
                reduced_frames_list = self._guessReduceFramesFromFile(
                    file_path=url.path()
                )
                for reduced_frames in reduced_frames_list:
                    self.addReduceFrames(reduced_frames)

    def supportedDropActions(self):
        """Inherited method to redefine supported drop actions."""
        return qt.Qt.CopyAction | qt.Qt.MoveAction

    def dragEnterEvent(self, event):
        if hasattr(event, "mimeData") and event.mimeData().hasFormat("text/uri-list"):
            event.accept()
            event.setDropAction(qt.Qt.CopyAction)
        else:
            try:
                qt.QListWidget.dragEnterEvent(self, event)
            except TypeError:
                pass

    def dragMoveEvent(self, event):
        if hasattr(event, "mimeDatamyitems") and event.mimeDatamyitems().hasFormat(
            "text/uri-list"
        ):
            event.setDropAction(qt.Qt.CopyAction)
            event.accept()
        else:
            try:
                qt.QListWidget.dragMoveEvent(self, event)
            except TypeError:
                pass


class ReduceDarkFlatSelectorWidget(qt.QSplitter):
    """
    Widget combining ReduceDarkFlatSelectorTableWidget and plot of the active item
    """

    sigUpdated = qt.Signal()
    """Signal emmit when the table is updated"""

    def __init__(self, parent: qt.QWidget | None = None) -> None:
        super().__init__(parent)
        self._first_display = True
        self._plot = Plot2D(self)
        self.addWidget(self._plot)
        self._table = ReduceDarkFlatSelectorTableWidget(parent=self)
        self.addWidget(self._table)

        # connect signal / slot
        self._table.sigActiveChanged.connect(self._updatePlot)
        self._table.sigUpdated.connect(self._tableUpdated)

    def _tableUpdated(self, *args, **kwargs):
        self.sigUpdated.emit()

    def _updatePlot(self, obj: numpy.ndarray | None):
        if obj is None:
            self._plot.clear()
        else:
            self._plot.addImage(
                data=obj,
                replace=True,
                resetzoom=self._first_display,
            )
            self._first_display = False

    # expose API
    def addReduceFrames(self, *args, **kwargs):
        self._table.addReduceFrames(*args, **kwargs)

    def getConfiguration(self) -> tuple:
        return self._table.getConfiguration()

    def setConfiguration(self, *args, **kwargs):
        self._table.setConfiguration(*args, **kwargs)

    def getSelectedReduceFrames(self) -> dict:
        return self._table.getSelectedReduceFrames()

    def clearSelection(self) -> None:
        self._table.clearSelection()

    def _removeSelected(self) -> None:
        self._table._removeSelected()

    def clear(self) -> None:
        self._table.clear()


class ReduceDarkFlatSelectorDialog(qt.QDialog):
    """
    'Final' dialog to select reduce frames
    """

    sigClearSelection = qt.Signal()
    """emit when user ask for the selection to be cleared"""
    sigSelectActiveAsDarks = qt.Signal(dict)
    """emit when user select the darks to be activated"""
    sigSelectActiveAsFlats = qt.Signal(dict)
    """emit when user select the flats to be activated"""

    sigUpdated = qt.Signal()
    """Signal emmit when the table is updated"""

    def __init__(self, parent=None) -> None:
        super().__init__(parent)

        self.setLayout(qt.QVBoxLayout())
        # add list
        self._widget = ReduceDarkFlatSelectorWidget()
        self.layout().addWidget(self._widget)
        # add buttons
        self._buttons = qt.QDialogButtonBox(parent=self)
        # select reduce flats
        self._selectAsFlatButton = qt.QPushButton(
            "Select as reduce flat(s)", parent=self
        )
        self._buttons.addButton(
            self._selectAsFlatButton, qt.QDialogButtonBox.AcceptRole
        )
        # select reduce darks
        self._selectAsDarkButton = qt.QPushButton(
            "Select as reduce dark(s)", parent=self
        )
        self._buttons.addButton(
            self._selectAsDarkButton, qt.QDialogButtonBox.AcceptRole
        )
        # clear
        self._clearSelectionButton = qt.QPushButton("clear selection", parent=self)
        self._buttons.addButton(
            self._clearSelectionButton, qt.QDialogButtonBox.ResetRole
        )
        # remove selected
        self._removeSelectionButton = qt.QPushButton("remove selection", parent=self)
        self._buttons.addButton(
            self._removeSelectionButton, qt.QDialogButtonBox.ActionRole
        )
        # remove selected
        self._addButton = qt.QPushButton("add", parent=self)
        self._buttons.addButton(self._addButton, qt.QDialogButtonBox.ActionRole)
        self.layout().addWidget(self._buttons)

        # connect signal / slot
        self._selectAsFlatButton.released.connect(self._flatSelected)
        self._selectAsDarkButton.released.connect(self._darkSelected)
        self._clearSelectionButton.released.connect(self._clearSelection)
        self._removeSelectionButton.released.connect(self._removeSelected)
        self._addButton.released.connect(self._addFromFileSelection)
        self._widget.sigUpdated.connect(self._tableUpdated)

    def _tableUpdated(self, *args, **kwargs):
        self.sigUpdated.emit()

    def _flatSelected(self):
        self.sigSelectActiveAsFlats.emit(
            self._widget.getSelectedReduceFrames(),
        )

    def _darkSelected(self):
        self.sigSelectActiveAsDarks.emit(
            self._widget.getSelectedReduceFrames(),
        )

    def _clearSelection(self):
        self._widget.clearSelection()

    def _removeSelected(self):
        self._widget._removeSelected()

    def _addFromFileSelection(self):
        dialog = DataFileDialog()
        dialog.setDirectory(get_default_directory())
        if dialog.exec_():
            url = dialog.selectedUrl()
            reduced_frames_tuple = ReduceDarkFlatSelectorTableWidget.get_reduce_frames(
                DataUrl(path=url)
            )
            for reduced_frames in reduced_frames_tuple:
                try:
                    self.addReduceFrames(reduced_frames)
                except Exception as e:
                    _logger.error(e)

    # expose API
    def addReduceFrames(self, *args, **kwargs):
        self._widget.addReduceFrames(*args, **kwargs)

    def getConfiguration(self) -> tuple:
        self._widget.getConfiguration()

    def setConfiguration(self, configuration: tuple) -> None:
        self._widget.setConfiguration(configuration)


def filter_selected_reduced_frames(configuration: tuple):
    """
    remove all the frames not 'selected' from the configuration.
    If a reduce frame set becomes empty it will also be removed

    :param configuration: configuration of the reduced frames
    """
    return _filter_reduced_frames(configuration=configuration, filter_selected=True)


def filter_unselected_reduced_frames(configuration: tuple):
    """
    remove all the frames 'selected' from the configuration.
    If a reduce frame set becomes empty it will also be removed

    :param configuration: configuration of the reduced frames
    """
    return _filter_reduced_frames(configuration=configuration, filter_selected=False)


def _filter_reduced_frames(configuration: tuple, filter_selected: bool):
    assert isinstance(
        configuration, (tuple, list, set)
    ), f"configuration is expected to be a tuple. Get {type(configuration)}"
    result = []
    for reduce_group in configuration:
        reduce_frames = reduce_group.get("reduce_frames", tuple())
        reduce_group_name = reduce_group.get("reduce_frames_name", None)
        res_set = tuple(
            filter(
                lambda my_dict: my_dict["selected"] == filter_selected,
                reduce_frames,
            )
        )
        if len(res_set) > 0:
            result_group = {}
            if reduce_group_name is not None:
                result_group["reduce_frames_name"] = reduce_group_name
            result_group["reduce_frames"] = res_set

            result.append(result_group)

    return tuple(result)



import numpy
from silx.utils.enum import Enum as _Enum


class Binning(_Enum):
    ONE_BY_ONE = "1x1"
    TWO_BY_TWO = "2x2"
    FOUR_BY_FOUR = "4x4"
    HEIGHT_BY_HEIGHT = "8x8"
    SIXTEEN_BY_SIXTEEN = "16x16"
    THIRTY_TWO_BY_THIRTY_TWO = "32x32"
    SIXTY_FOUR_BY_SIXTY_FOUR = "64x64"
    ONE_HUNDRED_TWENTY_HEIGHT_BY_ONE_HUNDRED_TWENTY_HEIGHT = "128x128"

    @staticmethod
    def _bin_data(data, binning):
        if not isinstance(data, numpy.ndarray):
            raise TypeError("data should be a numpy array")
        if not data.ndim == 2:
            raise ValueError("data is expected to be 2d")
        binning = Binning.from_value(binning)
        if binning is Binning.ONE_BY_ONE:
            return data
        elif binning is Binning.TWO_BY_TWO:
            return data[::2, ::2]
        elif binning is Binning.FOUR_BY_FOUR:
            return data[::4, ::4]
        elif binning is Binning.HEIGHT_BY_HEIGHT:
            return data[::8, ::8]
        elif binning is Binning.SIXTEEN_BY_SIXTEEN:
            return data[::16, ::16]
        elif binning is Binning.THIRTY_TWO_BY_THIRTY_TWO:
            return data[::32, ::32]
        elif binning is Binning.SIXTY_FOUR_BY_SIXTY_FOUR:
            return data[::64, ::64]
        else:
            raise NotImplementedError



# coding: utf-8
from __future__ import annotations

import logging
import os
import shutil
import tempfile
import time
import unittest

from silx.gui import qt
from silx.gui.utils.testutils import SignalListener, TestCaseQt

from tomwer.core.utils.scanutils import MockNXtomo
from tomwer.gui.visualization import sinogramviewer

logging.disable(logging.INFO)


class TestSinogramViewer(TestCaseQt):
    """unit test for the :class:_ImageStack widget"""

    def setUp(self):
        TestCaseQt.setUp(self)
        self._widget = sinogramviewer.SinogramViewer()
        self._widget.setAttribute(qt.Qt.WA_DeleteOnClose)

        self.tmp_dir = tempfile.mkdtemp()
        self.scan = MockNXtomo(
            scan_path=os.path.join(self.tmp_dir, "myscan"),
            n_proj=20,
            n_ini_proj=20,
            dim=10,
        ).scan

        # listen to the 'sigSinoLoadEnded' signal from the sinogram viewer
        self.signalListener = SignalListener()
        self._widget.sigSinoLoadEnded.connect(self.signalListener)

    def tearDown(self):
        shutil.rmtree(self.tmp_dir)
        self._widget.close()
        self._widget = None
        unittest.TestCase.tearDown(self)

    def test(self):
        """Make sur the addLeaf and clear functions are working"""
        self._widget.setScan(self.scan)
        timeout = 10
        while timeout >= 0 and self.signalListener.callCount() < 1:
            timeout -= 0.1
            time.sleep(0.1)
        if timeout >= 0:
            raise TimeoutError("widget never emitted the sigSinogramLoaded " "signal")
        self._widget.show()
        self.qWaitForWindowExposed(self._widget)
        self.qapp.processEvents()
        self.assertTrue(self._widget.getActiveImage() is not None)



# coding: utf-8
from __future__ import annotations

import gc
import os
import pickle
import tempfile

import h5py
import numpy
import pytest
from orangecanvas.scheme.readwrite import literal_dumps
from silx.gui import qt
from silx.gui.utils.testutils import TestCaseQt
from silx.io.url import DataUrl

from orangecontrib.tomwer.widgets.edit.DarkFlatPatchOW import DarkFlatPatchOW
from tomwer.tests.utils import skip_gui_test


@pytest.mark.skipif(skip_gui_test(), reason="skip gui test")
class TestDarkFlatPatch(TestCaseQt):
    """Test that the axis widget work correctly"""

    def setUp(self):
        super().setUp()
        self._window = DarkFlatPatchOW()

    def tearDown(self):
        self._window.setAttribute(qt.Qt.WA_DeleteOnClose)
        self._window.close()
        self._window = None
        gc.collect()

    def test(self):
        self._window.show()
        self.qWaitForWindowExposed(self._window)

    def test_serialiazation(self):
        with tempfile.TemporaryDirectory() as tmp_path:
            file_path = os.path.join(tmp_path, "darks.hdf5")
            with h5py.File(file_path, mode="w") as h5f:
                h5f.require_group("darks")["0"] = numpy.ones((10, 10))
            dark_url = DataUrl(file_path=file_path, data_path="darks/0")
            self._window.widget.setStartDarkUrl(url=dark_url)
            config = self._window.getConfiguration()
            pickle.dumps(config)
            literal_dumps(config)

            self._window.setConfiguration(config)



from __future__ import annotations

import fnmatch
import logging
import os

import h5py
from orangewidget import gui, widget
from orangewidget.widget import Output
from silx.gui import qt
from orangewidget.settings import Setting

import tomwer.core.process.control.datadiscovery
from orangecontrib.tomwer.widgets.utils import WidgetLongProcessing
from tomwer.core.scan.blissscan import BlissScan
from tomwer.core.scan.edfscan import EDFTomoScan
from tomwer.core.scan.nxtomoscan import NXtomoScan
from tomwer.core.scan.scanbase import TomwerScanBase
from tomwer.core.scan.scanfactory import ScanFactory
from tomwer.core.scan.scantype import ScanType
from tomwer.gui.control.datadiscovery import DataDiscoveryWidget

logger = logging.getLogger(__name__)


class DataDiscoveryOW(widget.OWBaseWidget, WidgetLongProcessing, openclass=True):
    """
    This widget will browse a folder and sub folder to find valid tomo scan project.
    Contrary to the scan watcher it will parse all folder / sub folders then stop.
    """

    name = "scan discovery"
    id = "orangecontrib.widgets.tomwer.control.DataDiscoveryOW.DataDiscoveryOW"
    description = (
        "This widget will browse a folder and sub folder to find valid tomo scan project. \n"
        "Contrary to the scan watcher it will parse all folder / sub folders then stop."
    )
    icon = "icons/datadiscover.svg"
    priority = 11
    keywords = [
        "tomography",
        "tomwer",
        "datadiscovery",
        "data",
        "discovery",
        "search",
        "research",
        "hdf5",
        "NXtomo",
    ]

    want_main_area = True
    want_control_area = False
    resizing_enabled = True

    ewokstaskclass = tomwer.core.process.control.datadiscovery._DataDiscoveryPlaceHolder

    class Outputs:
        data = Output(name="data", type=TomwerScanBase, doc="one scan to be process")

    settings = Setting(dict())

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._widget = DataDiscoveryWidget(parent=self)

        self._box = gui.vBox(self.mainArea, self.name)
        layout = self._box.layout()
        layout.addWidget(self._widget)

        self._widget.setConfiguration(self.settings)

        # getting ready for processing
        self._processingThread = _DiscoverThread()

        # connect signal / slot
        self._processingThread.sigDataFound.connect(self._sendData)
        self._processingThread.finished.connect(self._endProcessing)
        self._widget.widget.controlWidget._qpbstartstop.released.connect(
            self._switchObservation
        )
        # dedicated one for settings
        self._widget.widget.controlWidget._filterQLE.editingFinished.connect(
            self._updatesettings
        )
        self._widget.widget.controlWidget._qteFolderSelected.editingFinished.connect(
            self._updatesettings
        )
        self._widget.widget.configWidget.sigScanTypeChanged.connect(
            self._updatesettings
        )

    def _updatesettings(self):
        self.settings = self.getConfiguration()

    # expose some API
    def getConfiguration(self) -> dict:
        return self._widget.getConfiguration()

    def setConfiguration(self, config: dict):
        self._widget.setConfiguration(config)

    def getProcessingThread(self) -> qt.QThread:
        return self._processingThread

    def setFolderObserved(self, dir_: str):
        self._widget.widget.setFolderObserved(dir_)

    def setSearchScanType(self, scan_type):
        self._widget.widget.setSearchScanType(scan_type)

    def setFilePattern(self, pattern: str | None):
        self._widget.widget.setLinuxFilePattern(pattern)

    def _switchObservation(self, *args, **kwargs):
        """stop or start the disceovery according to the thread state"""
        thread = self.getProcessingThread()
        if thread.isRunning():
            thread.quit()
            self._endProcessing()
        else:
            self.start_discovery()

    def start_discovery(self, wait: int | None = None):
        """start the discovery of scans
        :param wait: optional waiting time in second
        """
        thread = self.getProcessingThread()
        if thread.isRunning():
            logger.warning("Discovery is already running")
            return
        self._startProcessing()
        thread.setConfiguration(self.getConfiguration())
        thread.start()
        if wait is not None:
            thread.wait(wait)

    def _sendData(self, data):
        if data is not None:
            self.Outputs.data.send(data)

    def _startProcessing(self, *args, **kwargs):
        self._widget.widget.controlWidget._qpbstartstop.setText("stop discovery")
        return super()._startProcessing(*args, **kwargs)

    def _endProcessing(self, *args, **kwargs):
        self._widget.widget.controlWidget._qpbstartstop.setText("start discovery")
        return super()._endProcessing(*args, **kwargs)


class _DiscoverThread(qt.QThread):
    """
    Thread to browse folder and sub folder and looking for some scan
    """

    sigDataFound = qt.Signal(object)
    """emit each time a new dataset is found"""

    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)
        self._root_dir = None
        self._research_type = None
        self._file_filter = None
        self._look_for_hdf5 = False

    def setConfiguration(self, config: dict):
        self._root_dir = config.get("start_folder", None)
        self._research_type = ScanType.from_value(config["scan_type_searched"])
        self._file_filter = config.get("file_filter", None)
        self._look_for_hdf5 = self._research_type in (
            ScanType.BLISS,
            ScanType.NX_TOMO,
        )

    def run(self):
        if self._research_type is None:
            logger.error("No reserach type defined. Cannot research scan")
        elif not os.path.isdir(self._root_dir):
            logger.error(f"{self._root_dir} is not a directory")
        else:
            self.discover_scan(file_path=self._root_dir)

    def discover_scan(self, file_path):
        if (
            os.path.isfile(file_path)
            and self._look_for_hdf5
            and h5py.is_hdf5(file_path)
        ) or (os.path.isdir(file_path) and not self._look_for_hdf5):
            try:
                name_match = self._file_filter is None or fnmatch.fnmatch(
                    os.path.basename(file_path), self._file_filter
                )
            except Exception as e:
                logger.error(f"fnmatch fail. Error is {e}")
                name_match = True
            if name_match:
                self._treat_path(file_path)
        if os.path.isdir(file_path):
            [
                self.discover_scan(file_path=os.path.join(file_path, file_))
                for file_ in os.listdir(file_path)
            ]

    def _treat_path(self, folder_path: str):
        """treat file / folder at a specific location"""
        try:
            scan_objs = ScanFactory.create_scan_objects(
                scan_path=folder_path,
                accept_bliss_scan=(self._research_type == ScanType.BLISS),
            )
        except Exception as e:
            logger.info(f"Fail to treat {folder_path}. Error is {e}")
        else:
            for scan in scan_objs:
                if (
                    isinstance(scan, BlissScan)
                    and self._research_type is ScanType.BLISS
                ):
                    self.sigDataFound.emit(scan)
                elif (
                    isinstance(scan, NXtomoScan)
                    and self._research_type is ScanType.NX_TOMO
                ):
                    self.sigDataFound.emit(scan)
                elif (
                    isinstance(scan, EDFTomoScan)
                    and self._research_type is ScanType.SPEC
                ):
                    self.sigDataFound.emit(scan)
                else:
                    raise NotImplementedError("case not handled", self._research_type)



# coding: utf-8
from __future__ import annotations

import shutil
import tempfile
import unittest

from tomoscan.io import HDF5File, get_swmr_mode

from tomwer.core.process.task import Task
from tomwer.core.utils.scanutils import MockNXtomo


class TestProcessRegistration(unittest.TestCase):
    """
    Make sure utils link to the process registration are
    correctly working
    """

    class DummyProcess(Task):
        @staticmethod
        def program_name():
            """Name of the program used for this processing"""
            return "dummy program"

        @staticmethod
        def program_version():
            """version of the program used for this processing"""
            return "0.0.0"

        @staticmethod
        def definition():
            """definition of the process"""
            return "no definition"

    def setUp(self):
        self.tmp_dir = tempfile.mkdtemp()
        self.scan = MockNXtomo(scan_path=self.tmp_dir, n_proj=2).scan

    def tearDown(self):
        shutil.rmtree(self.tmp_dir)

    def testGetProcessNodes(self):
        """insure it return the last dark process based on the processing index"""

        for i in range(20):
            Task._register_process(
                self.scan.process_file,
                process=self.DummyProcess,
                entry=self.scan.entry,
                configuration=None,
                results={"output": i},
                process_index=i,
            )

        with HDF5File(self.scan.process_file, "r", swmr=get_swmr_mode()) as h5f:
            nodes = Task._get_process_nodes(
                root_node=h5f[self.scan.entry], process=self.DummyProcess
            )
            self.assertEqual(len(nodes), 20)
            self.assertTrue("/entry/tomwer_process_16" in nodes)
            self.assertEqual(nodes["/entry/tomwer_process_16"], 16)
            self.assertTrue("/entry/tomwer_process_1" in nodes)
            self.assertEqual(nodes["/entry/tomwer_process_1"], 1)



# coding: utf-8

from __future__ import annotations

import copy
import logging
from contextlib import AbstractContextManager

from orangewidget import gui
from orangewidget.settings import Setting
from orangewidget.widget import Input, Output
from silx.gui import qt

import tomwer.core.process.reconstruction.nabu.nabuslices
from tomwer.core.cluster import SlurmClusterConfiguration
from tomwer.core.futureobject import FutureTomwerObject
from tomwer.core.process.reconstruction.nabu.nabuslices import NabuSlicesTask
from tomwer.core.scan.scanbase import TomwerScanBase, _TomwerBaseDock
from tomwer.core.volume.volumefactory import VolumeFactory

from tomwer.gui.reconstruction.nabu.slices import NabuWindow
from tomwer.synctools.stacks.reconstruction.nabu import NabuSliceProcessStack

from ...orange.managedprocess import SuperviseOW
from ..utils import WidgetLongProcessing

_logger = logging.getLogger(__name__)


class NabuOW(WidgetLongProcessing, SuperviseOW):
    """
    A simple widget managing the copy of an incoming folder to an other one

    :param parent: the parent widget
    """

    # note of this widget should be the one registered on the documentation
    name = "nabu slice reconstruction"
    id = "orange.widgets.tomwer.reconstruction.NabuOW.NabuOW"
    description = "This widget will call nabu for running a reconstruction "
    icon = "icons/nabu_2d.svg"
    priority = 12
    keywords = ["tomography", "nabu", "reconstruction", "FBP", "filter"]

    want_main_area = True
    resizing_enabled = True

    _ewoks_default_inputs = Setting({"data": None, "nabu_params": None})

    ewokstaskclass = tomwer.core.process.reconstruction.nabu.nabuslices.NabuSlicesTask

    sigScanReady = qt.Signal(TomwerScanBase)
    "Signal emitted when a scan is ended"

    class Inputs:
        reprocess = Input(
            name="change recons params",
            type=_TomwerBaseDock,
            doc="reconpute slice with different parameters",
        )
        data = Input(
            name="data",
            type=TomwerScanBase,
            doc="one scan to be process",
            default=True,
            multiple=True,
        )
        cluster_in = Input(
            name="cluster_config",
            type=SlurmClusterConfiguration,
            doc="slurm cluster to be used",
            multiple=False,
        )

    class Outputs:
        data = Output(name="data", type=TomwerScanBase, doc="one scan to be process")

        future_out = Output(
            name="future_tomo_obj",
            type=FutureTomwerObject,
            doc="data with some remote processing",
        )

        slice_urls = Output(name="slice urls", type=tuple, doc="tuple of urls created")

    class DialogCM(AbstractContextManager):
        """Simple context manager to hide / show button dialogs"""

        def __init__(self, dialogButtonsBox):
            self._dialogButtonsBox = dialogButtonsBox

        def __enter__(self):
            self._dialogButtonsBox.show()

        def __exit__(self, exc_type, exc_val, exc_tb):
            self._dialogButtonsBox.hide()

    def __init__(self, parent=None):
        SuperviseOW.__init__(self, parent)
        WidgetLongProcessing.__init__(self)
        self._slurmCluster = None
        self.__exec_for_ci = False
        # processing tool
        self._processingStack = NabuSliceProcessStack(self, process_id=self.process_id)

        _layout = gui.vBox(self.mainArea, self.name).layout()
        # main widget
        self._nabuWidget = NabuWindow(parent=self)
        _layout.addWidget(self._nabuWidget)
        # add button to validate when change reconstruction parameters is
        # called
        types = qt.QDialogButtonBox.Ok
        self._buttons = qt.QDialogButtonBox(self)
        self._buttons.setStandardButtons(types)
        _layout.addWidget(self._buttons)

        # set up
        self._buttons.hide()

        # load settings
        nabu_params = self._ewoks_default_inputs.get("nabu_params", None)
        if nabu_params not in (dict(), None):
            try:
                self._nabuWidget.setConfiguration(nabu_params)
            except Exception:
                _logger.warning("fail to load reconstruction settings")

        # expose API
        self.getMode = self._nabuWidget.getMode
        self.setMode = self._nabuWidget.setMode

        # connect signal / slot
        self._processingStack.sigComputationStarted.connect(self._startProcessing)
        self._processingStack.sigComputationEnded.connect(self._endProcessing)
        self._buttons.button(qt.QDialogButtonBox.Ok).clicked.connect(self.accept)
        self._nabuWidget.sigConfigChanged.connect(self._updateSettingsVals)

    def __new__(cls, *args, **kwargs):
        # ensure backward compatibility with 'static_input'
        static_input = kwargs.get("stored_settings", {}).get("static_input", None)
        if static_input not in (None, {}):
            _logger.warning(
                "static_input has been deprecated. Will be replaced by _ewoks_default_inputs in the workflow file. Please save the workflow to apply modifications"
            )
            kwargs["stored_settings"]["_ewoks_default_inputs"] = static_input
        return super().__new__(cls, *args, **kwargs)

    @Inputs.data
    def process(self, scan, *args, **kwargs):
        assert isinstance(scan, (TomwerScanBase, type(None)))
        if scan is None:
            return
        scan_ = copy.copy(scan)
        scan_.clear_latest_reconstructions()

        _logger.info(f"add {scan} to the stack")
        # update the reconstruction mode if possible
        self._nabuWidget.setScan(scan_)
        self._processingStack.add(scan_, self.getConfiguration())

    @Inputs.reprocess
    def reprocess(self, scan):
        """Recompute nabu with different parameters"""
        # wait for user to tune the reconstruction
        if scan is None:
            return

        if scan.axis_params is None or scan.axis_params.relative_cor_value is None:
            # try to retrieve last computed cor value from nabu process
            r_cor = NabuSlicesTask.retrieve_last_relative_cor(scan)
            if scan.axis_params is None:
                from tomwer.synctools.axis import QAxisRP

                scan.axis_params = QAxisRP()
            scan.axis_params.set_relative_value(float(r_cor))

        self.show()
        with NabuOW.DialogCM(self._buttons):
            if self.__exec_for_ci is True:
                self._ciExec()
            else:
                if self.exec_():
                    # for now The behavior for reprocessing is the sama as for processing
                    if hasattr(scan, "instance"):
                        self.process(scan.instance)
                    else:
                        self.process(scan)

    def cancel(self, scan):
        if scan is None:
            return
        if scan in self._processingStack:
            self._processingStack.remove(scan)
        if (
            self._processingStack._data_currently_computed.get_identifier().to_str()
            == scan.get_identifier().to_str()
        ):
            # stop current processing
            self._processingStack.cancel()
            # if possible process next
            if self._processingStack.can_process_next():
                self._processingStack._process_next()

    @Inputs.cluster_in
    def setCluster(self, slurm_cluster: SlurmClusterConfiguration | None):
        assert isinstance(
            slurm_cluster, (type(None), SlurmClusterConfiguration)
        ), f"Expect None of SlurmClusterConfiguration. Not {type(slurm_cluster)}"
        self._slurmCluster = slurm_cluster

    def _endProcessing(self, scan, future_tomo_obj):
        WidgetLongProcessing._endProcessing(self, scan)
        if scan is not None:
            # send scan
            self.Outputs.data.send(scan)
            # send slice urls
            slice_urls = []
            for rec_identifier in scan.latest_reconstructions:
                slice_urls.extend(
                    VolumeFactory.from_identifier_to_vol_urls(rec_identifier)
                )
            if len(slice_urls) > 0:
                slice_urls = tuple(slice_urls)
                # provide list of reconstructed slices
                self.Outputs.slice_urls.send(slice_urls)

            self.sigScanReady.emit(scan)
        if future_tomo_obj is not None:
            # send future scan
            self.Outputs.future_out.send(future_tomo_obj)

    def setDryRun(self, dry_run):
        self._processingStack.setDryRun(dry_run)

    def _ciExec(self):
        self.activateWindow()
        self.raise_()
        self.show()

    def _replaceExec_(self):
        """used for CI, replace the exec_ call ny"""
        self.__exec_for_ci = True

    def _updateSettingsVals(self):
        self._ewoks_default_inputs = {
            "data": None,
            "nabu_params": self.getConfiguration(),
        }

    def getConfiguration(self):
        config = self._nabuWidget.getConfiguration()
        config["cluster_config"] = self._slurmCluster
        return config

    def setConfiguration(self, config):
        # ignore slurm cluster. Defined by the upper widget
        config.pop("cluster_config", None)
        self._nabuWidget.setConfiguration(config=config)

    def keyPressEvent(self, event):
        """The event has to be filtered since we have some children
        that can be edited using the 'enter' key as defining the cor manually
        (see #481)). As we are in a dialog this automatically trigger
        'accepted'. See https://forum.qt.io/topic/5080/preventing-enter-key-from-triggering-ok-in-qbuttonbox-in-particular-qlineedit-qbuttonbox/5
        """
        if event.key() not in (qt.Qt.Key_Enter, qt.Qt.Key_Return):
            super().keyPressEvent(event)



# coding: utf-8
from __future__ import annotations


import shutil
import tempfile
import unittest

from tomwer.core.process.control.timer import TimerTask
from tomwer.core.scan.scanbase import TomwerScanBase
from tomwer.core.utils.scanutils import MockEDF


class TestTimerIO(unittest.TestCase):
    """Test inputs and outputs types of the handler functions"""

    def setUp(self):
        self.scan_folder = tempfile.mkdtemp()

        self.scan = MockEDF.mockScan(
            scanID=self.scan_folder, nRadio=10, nRecons=1, nPagRecons=4, dim=10
        )

    def tearDown(self):
        shutil.rmtree(self.scan_folder)

    def testInputOutput(self):
        """Test that io using TomoBase instance work"""
        for input_type in (dict, TomwerScanBase):
            for serialize_output_data in (True, False):
                with self.subTest(
                    return_dict=serialize_output_data,
                    input_type=input_type,
                ):
                    input_obj = self.scan
                    if input_obj is dict:
                        input_obj = input_obj.to_dict()

                    timer_process = TimerTask(
                        inputs={
                            "wait": 0.1,
                            "serialize_output_data": serialize_output_data,
                            "data": input_obj,
                        }
                    )
                    timer_process.run()
                    out = timer_process.outputs.data
                    if serialize_output_data:
                        self.assertTrue(isinstance(out, dict))
                    else:
                        self.assertTrue(isinstance(out, TomwerScanBase))



# coding: utf-8
from __future__ import annotations

import logging

from orangewidget import gui, settings, widget
from orangewidget.widget import Input
from silx.gui import qt

import tomwer.core.process.visualization.dataviewer
from tomwer.core.scan.scanbase import TomwerScanBase
from tomwer.gui import icons, utils
from tomwer.gui.visualization.dataviewer import DataViewer

_logger = logging.getLogger(__name__)


class DataViewerOW(widget.OWBaseWidget, openclass=True):
    """a data viewer able to:

    - display slices (latest reconstructed if any)
    - display radios with or without normalization

    :param parent: the parent widget
    """

    name = "data viewer"
    id = "orange.widgets.tomwer.dataviewer"
    description = "allow user too browse through data"
    icon = "icons/eye.png"
    priority = 70
    keywords = ["tomography", "file", "tomwer", "acquisition", "validation"]

    want_main_area = True
    want_control_area = False
    resizing_enabled = True

    ewokstaskclass = tomwer.core.process.visualization.dataviewer._DataViewerPlaceHolder

    _viewer_config = settings.Setting(dict())

    class Inputs:
        data = Input(
            name="data",
            type=TomwerScanBase,
            multiple=True,
        )

    def __init__(self, parent=None):
        super().__init__(parent)
        self._imagejThreads = []
        # threads used to open frames with image j
        self._layout = gui.vBox(self.mainArea, self.name).layout()
        self.viewer = DataViewer(parent=self)
        self.viewer.setSizePolicy(qt.QSizePolicy.Expanding, qt.QSizePolicy.Expanding)
        self._layout.addWidget(self.viewer)
        self._setSettings(settings=self._viewer_config)

        # open with ImageJ button
        types = qt.QDialogButtonBox.Apply
        self._buttons = qt.QDialogButtonBox(self)
        self._buttons.setStandardButtons(types)
        self._buttonOpenImageJ = self._buttons.button(qt.QDialogButtonBox.Apply)
        self._buttonOpenImageJ.setText("open with ImageJ")
        self._buttonOpenImageJ.setIcon(icons.getQIcon("Imagej_icon"))
        self.layout().addWidget(self._buttons)

        # connect signal / slot
        self.viewer.sigConfigChanged.connect(self._updateSettings)
        self._buttonOpenImageJ.released.connect(self._openCurrentInImagej)

    def _openCurrentInImagej(self):
        current_url = self.viewer.getCurrentUrl()

        if current_url is None:
            _logger.warning("No active image defined")
        else:
            try:
                self._imagejThreads.append(utils.open_url_with_image_j(current_url))
            except OSError as e:
                msg = qt.QMessageBox(self)
                msg.setIcon(qt.QMessageBox.Warning)
                msg.setWindowTitle("Unable to open image in imagej")
                msg.setText(str(e))
                msg.exec_()

    @Inputs.data
    def addScan(self, scan, *args, **kwargs):
        if scan is None:
            return
        self.viewer.setScan(scan)

    def sizeHint(self):
        return qt.QSize(400, 500)

    def _updateSettings(self):
        self._viewer_config["mode"] = (  # pylint: disable=E1137
            self.viewer.getDisplayMode().value
        )
        self._viewer_config["slice_opt"] = (  # pylint: disable=E1137
            self.viewer.getSliceOption().value
        )
        self._viewer_config["radio_opt"] = (  # pylint: disable=E1137
            self.viewer.getRadioOption().value
        )

    def _setSettings(self, settings):
        old_state = self.viewer.blockSignals(True)
        if "mode" in settings:
            self.viewer.setDisplayMode(settings["mode"])
        if "slice_opt" in settings:
            self.viewer.setSliceOption(settings["slice_opt"])
        if "radio_opt" in settings:
            self.viewer.setRadioOption(settings["radio_opt"])
        self.viewer.blockSignals(old_state)

    def close(self):
        [thread.quit() for thread in self._imagejThreads]
        self.viewer.close()
        self.viewer = None
        super().close()



# coding: utf-8
from __future__ import annotations


import logging

from tomwer.core.scan.scanbase import TomwerScanBase
from tomwer.gui.control.datalist import GenericScanList
from tomwer.gui.dialog.QDataDialog import QDataDialog

from .selectorwidgetbase import _SelectorWidget

logger = logging.getLogger(__name__)


class ScanSelectorWidget(_SelectorWidget):
    """Widget used to select a scan on a list"""

    def _buildDataList(self):
        return GenericScanList(parent=self)

    def n_scan(self) -> int:
        return super().n_data()

    def _callbackAddData(self):
        dialog = QDataDialog(self, multiSelection=True)

        if not dialog.exec_():
            dialog.close()
            return

        for folder in dialog.files_selected():
            tomo_objs = self.add(folder)
            self.setMySelection(tomo_objs)
        self.sigUpdated.emit()

    def setActiveData(self, data):
        if isinstance(data, TomwerScanBase):
            data = data.path
        super().setActiveData(data)



from __future__ import annotations

from tomwer.io.utils.utils import str_to_dict

from nabu.stitching import config as stitching_config
from nabu.stitching.overlap import OverlapStitchingStrategy
from silx.gui import qt
from tomwer.gui.utils.scrollarea import QComboBoxIgnoreWheel as QComboBox
from tomwer.gui.utils.scrollarea import QSpinBoxIgnoreWheel as QSpinBox
from ..singleaxis import _SingleAxisMixIn


class _StitchingHeightSpinBox(qt.QWidget):
    def __init__(self, parent=None) -> None:
        super().__init__(parent)
        self.setLayout(qt.QHBoxLayout())
        self._autoCB = qt.QCheckBox("max", self)
        self.layout().addWidget(self._autoCB)
        self._stitchingHeight = QSpinBox(self)
        self._stitchingHeight.setMinimum(3)
        self._stitchingHeight.setMaximum(999999)
        self._stitchingHeight.setValue(40)
        self._stitchingHeight.setSuffix("px")
        self.layout().addWidget(self._stitchingHeight)

        self._autoCB.toggled.connect(self._stitchingHeight.setDisabled)

    def getStitchingHeight(self) -> int | None:
        if self._autoCB.isChecked():
            return None
        else:
            return self._stitchingHeight.value()

    def setStitchingHeight(self, height: int | None) -> None:
        self._autoCB.setChecked(height is None)
        if height is not None:
            self._stitchingHeight.setValue(int(height))


class StitchingStrategies(qt.QWidget, _SingleAxisMixIn):
    """
    Defines algorithm and strategies to be used
    """

    DEFAULT_STITCHING_HEIGHT = None  # max value is the default

    sigChanged = qt.Signal()
    """emit when strategy change"""

    def __init__(self, axis: int, parent=None) -> None:
        super().__init__(parent)
        self.setLayout(qt.QFormLayout())
        self._stitchingStrategyCG = QComboBox(parent=self)
        for strategy in OverlapStitchingStrategy.values():
            self._stitchingStrategyCG.addItem(strategy)
        self._stitchingStrategyCG.setToolTip(
            "stitcher behavior is also know as stitching strategy. It define the behavior to have on overlaping areas"
        )
        self.layout().addRow("stitcher behavior", self._stitchingStrategyCG)
        self._axis = axis  # needed to implement the _SingleAxisMixIn interface

        self._stitchingHeight = _StitchingHeightSpinBox(parent=self)
        self._stitchingHeight.setStitchingHeight(self.DEFAULT_STITCHING_HEIGHT)
        self.layout().addRow("stitching height", self._stitchingHeight)

        # set up
        idx = self._stitchingStrategyCG.findText(
            OverlapStitchingStrategy.COSINUS_WEIGHTS.value
        )
        self._stitchingStrategyCG.setCurrentIndex(idx)

        # connect signal / slot
        self._stitchingStrategyCG.currentIndexChanged.connect(self._changed)

    def _changed(self, *args, **kwargs):
        self.sigChanged.emit()

    def getStitchingStrategy(self) -> OverlapStitchingStrategy:
        return OverlapStitchingStrategy.from_value(
            self._stitchingStrategyCG.currentText()
        )

    def setStitchingStrategy(self, strategy: OverlapStitchingStrategy | str):
        strategy = OverlapStitchingStrategy.from_value(strategy)
        idx = self._stitchingStrategyCG.findText(strategy.value)
        if idx >= 0:
            self._stitchingStrategyCG.setCurrentIndex(idx)

    def getConfiguration(self) -> dict:
        overlap_size = self._stitchingHeight.getStitchingHeight()
        if overlap_size is None:
            overlap_size = ""
        return {
            stitching_config.STITCHING_SECTION: {
                stitching_config.STITCHING_STRATEGY_FIELD: self.getStitchingStrategy().value,
                f"axis_{self.first_axis}_params": f"overlap_size={overlap_size}",
            }
        }

    def setConfiguration(self, config: dict):
        strategy = config.get(stitching_config.STITCHING_SECTION, {}).get(
            stitching_config.STITCHING_SECTION, None
        )
        if strategy is not None:
            self.setStitchingStrategy(strategy=strategy)

        first_axis_params_dict = str_to_dict(
            config.get(stitching_config.STITCHING_SECTION, {}).get(
                f"axis_{self._axis}_params", ""
            )
        )
        stitching_height = first_axis_params_dict.get(
            stitching_config.KEY_OVERLAP_SIZE, "unknown"
        )
        if stitching_height in ("None", "", None):
            self._stitchingHeight.setStitchingHeight(None)
        elif stitching_height != "unknown":
            self._stitchingHeight.setStitchingHeight(stitching_height)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import argparse
import logging
import os
import signal
import sys

from silx.gui import qt

from tomwer.core.process.reconstruction.normalization import SinoNormalizationTask
from tomwer.core.scan.nxtomoscan import NXtomoScan
from tomwer.core.scan.scanbase import TomwerScanBase
from tomwer.core.scan.scanfactory import ScanFactory
from tomwer.gui import icons
from tomwer.gui.reconstruction.normalization.intensity import (
    SinoNormWindow as _SinoNormWindow,
)
from tomwer.gui.utils.splashscreen import getMainSplashScreen

logging.basicConfig(level=logging.INFO)
_logger = logging.getLogger(__name__)


class IntensityNormalizationThread(qt.QThread):
    """
    Thread to call the normalization process when needed (to avoid gui freeze)
    """

    def __init__(self, *args, **kwargs):
        qt.QThread.__init__(self, *args, **kwargs)
        self._result = None

    @property
    def result(self):
        return self._result

    def init(self, data, configuration):
        if not isinstance(configuration, dict):
            raise TypeError("Configuration is expected to be a dict")
        if not isinstance(data, TomwerScanBase):
            raise TypeError("Scan is expected to be an instance of " "TomwerScanBase")
        self.scan = data
        self._configuration = configuration
        self._result = None

    def run(self) -> None:
        process = SinoNormalizationTask(
            process_id=None,
            inputs={
                "data": self.scan,
                "configuration": self._configuration,
                "serialize_output_data": False,
            },
            varinfo=None,
        )
        process.run()
        self._result = self.scan.intensity_normalization.tomwer_processing_res


class NormIntensityWindow(_SinoNormWindow):
    def __init__(self, parent=None):
        _SinoNormWindow.__init__(self, parent)
        self._processingThread = IntensityNormalizationThread()
        self._hideLockButton()

        # connect signal / slot
        self._optsWidget.sigProcessingRequested.connect(self._launchProcessing)
        self._processingThread.finished.connect(self._threadedProcessEnded)

    def _validated(self):
        scan = self.getScan()
        if scan is not None:
            with scan.acquire_process_file_lock():
                SinoNormalizationTask._register_process(
                    process_file=scan.process_file,
                    process=SinoNormalizationTask,
                    entry=scan.entry,
                    configuration=self.getConfiguration(),
                    results={
                        "method": scan.intensity_normalization.method.value,
                        "extra_infos": scan.intensity_normalization.get_extra_infos(),
                    },
                    process_index=scan.pop_process_index(),
                )
        self.close()

    def _launchProcessing(self):
        self.clear()
        if self._processingThread.isRunning():
            _logger.error(
                "a calculation is already launch. You must wait for "
                "it to end prior to launch a new one"
            )
        else:
            qt.QApplication.setOverrideCursor(qt.Qt.WaitCursor)
            self._processingThread.init(
                data=self.getScan(),
                configuration=self.getConfiguration(),
            )
            self._processingThread.start()

    def _threadedProcessEnded(self):
        qt.QApplication.restoreOverrideCursor()

    def close(self) -> None:
        self._stopProcessingThread()
        super().close()

    def _stopProcessingThread(self):
        if self._processingThread:
            self._processingThread.terminate()
            self._processingThread.wait(500)
            self._processingThread = None

    def stop(self):
        self._stopProcessingThread()
        super().stop()


def main(argv):
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument(
        "scan_path",
        help="For EDF acquisition: provide folder path, for HDF5 / nexus"
        "provide the master file",
        default=None,
    )
    parser.add_argument(
        "entry",
        help="For Nexus files: entry in the master file",
        default=None,
        nargs="?",
    )
    parser.add_argument(
        "--debug",
        dest="debug",
        action="store_true",
        default=False,
        help="Set logging system in debug mode",
    )

    options = parser.parse_args(argv[1:])

    if options.debug:
        logging.root.setLevel(logging.DEBUG)

    global app  # QApplication must be global to avoid seg fault on quit
    app = qt.QApplication.instance() or qt.QApplication(["tomwer"])
    splash = getMainSplashScreen()
    qt.QApplication.setOverrideCursor(qt.Qt.WaitCursor)
    qt.QApplication.processEvents()

    qt.QLocale.setDefault(qt.QLocale(qt.QLocale.English))
    qt.QLocale.setDefault(qt.QLocale.c())
    signal.signal(signal.SIGINT, sigintHandler)
    sys.excepthook = qt.exceptionHandler

    timer = qt.QTimer()
    timer.start(500)
    # Application have to wake up Python interpreter, else SIGINT is not
    # catch
    timer.timeout.connect(lambda: None)

    if options.scan_path is not None:
        if os.path.isdir(options.scan_path):
            options.scan_path = options.scan_path.rstrip(os.path.sep)
            scan = ScanFactory.create_scan_object(scan_path=options.scan_path)
        else:
            if not os.path.isfile(options.scan_path):
                raise ValueError(
                    "scan path should be a folder containing an"
                    " EDF acquisition or an hdf5 - nexus "
                    "compliant file"
                )
            if options.entry is None:
                raise ValueError("entry in the master file should be specify")
            scan = NXtomoScan(scan=options.scan_path, entry=options.entry)
    else:
        scan = ScanFactory.mock_scan()
    # define the process_index is any tomwer_processes_existing
    if options.debug:
        _logger.setLevel(logging.DEBUG)

    window = NormIntensityWindow(parent=None)

    window.setWindowTitle("sinogram-norm-intensity")
    window.setWindowIcon(icons.getQIcon("tomwer"))
    window.setScan(scan)

    splash.finish(window)
    window.show()
    qt.QApplication.restoreOverrideCursor()
    app.aboutToQuit.connect(window.stop)
    app.exec_()


def getinputinfo():
    return "tomwer intensity normalization [scanDir entry]"


def sigintHandler(*args):
    """Handler for the SIGINT signal."""
    qt.QApplication.quit()


if __name__ == "__main__":
    main(sys.argv)



"""Application to define a tomography workflow (also as know as 'ewoks-canvas' or 'orange canvas')
"""

import logging
import platform
import sys

import psutil

from tomwer.app.canvas_launcher.mainwindow import OMain as QMain
from tomwer.app.canvas_launcher.utils import get_tomotools_stack_versions
from tomwer.core.utils.resource import increase_max_number_file

_logger = logging.getLogger(__name__)


def print_versions():
    """Prints the Tomotools packages versions"""
    print(
        "\n - ".join(
            [
                "Tomotools packages versions:",
            ]
            + list(
                [
                    f"{soft_name: <11}: {version}"
                    for soft_name, version in get_tomotools_stack_versions().items()
                ]
            )
        )
    )


def check_free_space():
    """Check if enough space exists. Else could bring issues on processing"""
    if platform.system() == "Linux":
        try:
            free_home_space = psutil.disk_usage("/home").free
        except OSError:
            pass
        else:
            if free_home_space < 100 * 1024 * 1024:
                # if no space Qt might fail to create the display and log file not be created
                _logger.warning(
                    f"only {free_home_space / 1024} ko available. Display might fail"
                )


def main(argv=None):
    print_versions()
    check_free_space()
    increase_max_number_file()
    return QMain().run(argv)


if __name__ == "__main__":
    sys.exit(main())



# coding: utf-8
"""
contains utils for inputs and outputs
"""

from __future__ import annotations

import os

import h5py
import numpy
from silx.io.url import DataUrl
from tomoscan.io import HDF5File

from tomwer.core.utils import scanutils
from tomwer.io.utils import (
    get_default_directory,
    get_linked_files_with_entry,
    get_slice_data,
)


def test_get_linked_files_with_entry(tmp_path):
    """test get_linked_files_with_entry function"""
    dir_test = tmp_path / "sub"
    dir_test.mkdir()

    layout = h5py.VirtualLayout(shape=(4, 10, 10), dtype="i4")
    for i_file in range(4):
        file_name = os.path.join(dir_test, f"file_{i_file}.hdf5")
        with HDF5File(file_name, "w") as h5s:
            h5s.create_dataset("data", (10, 10), "i4", numpy.ones((10, 10)))
            vsource = h5py.VirtualSource(file_name, "data", shape=(10, 10))
            layout[i_file] = vsource

    test_with_vds = os.path.join(dir_test, "file_with_vds")
    with HDF5File(test_with_vds, "w") as h5s:
        h5s.create_virtual_dataset("vdata", layout, fillvalue=-1)

    assert len(get_linked_files_with_entry(test_with_vds, "vdata")) == 4


def test_get_default_directory():
    get_default_directory()


def test_get_slice_data_vol(tmp_path):
    """test load of a .vol file"""
    dir_test = tmp_path / "test_vol"
    dir_test.mkdir()
    vol_file_path = os.path.join(dir_test, "volume.vol")
    vol_info_file_path = os.path.join(dir_test, "volume.vol.info")

    shape = (4, 50, 20)
    data = numpy.ones(shape)
    data.astype(numpy.float32).tofile(vol_file_path)
    scanutils.MockEDF._createVolInfoFile(
        filePath=vol_info_file_path,
        shape=shape,
    )

    vol = get_slice_data(DataUrl(file_path=vol_file_path))
    assert vol is not None
    assert vol.shape == shape
    vol = get_slice_data(DataUrl(file_path=vol_info_file_path))
    assert vol is not None
    assert vol.shape == shape



from tomwer.core.process.drac.gallery import (
    deduce_dataset_gallery_location,
    deduce_proposal_GALLERY_location,
    PROPOSAL_GALLERY_DIR_NAME,
)
from tomwer.core.scan.edfscan import EDFTomoScan
from tomwer.core.scan.nxtomoscan import NXtomoScan


def test_deduce_dataset_gallery_dir():
    """test the deduce_gallery_dir function"""
    assert (
        deduce_dataset_gallery_location(
            NXtomoScan(scan="/path/to/PROCESSED_DATA/my_scan.nx", entry="entry")
        )
        == "/path/to/PROCESSED_DATA/gallery"
    )
    assert (
        deduce_dataset_gallery_location(
            NXtomoScan(scan="/path/to/RAW_DATA/collection/my_scan.nx", entry="entry")
        )
        == "/path/to/PROCESSED_DATA/collection/gallery"
    )
    assert (
        deduce_dataset_gallery_location(
            NXtomoScan(scan="/any/random/path/my_scan.nx", entry="entry")
        )
        == "/any/random/path/gallery"
    )
    assert (
        deduce_dataset_gallery_location(
            EDFTomoScan(scan="/path/to/PROCESSED_DATA/dataset/toto")
        )
        == "/path/to/PROCESSED_DATA/dataset/toto/gallery"
    )
    assert (
        deduce_dataset_gallery_location(EDFTomoScan(scan="/path/to/dataset/toto"))
        == "/path/to/dataset/toto/gallery"
    )


def test_deduce_proposal_gallery_dir():
    """test the deduce_gallery_dir function"""
    assert (
        deduce_proposal_GALLERY_location(
            NXtomoScan(scan="/path/to/PROCESSED_DATA/my_scan.nx", entry="entry")
        )
        == f"/path/to/{PROPOSAL_GALLERY_DIR_NAME}"
    )
    assert (
        deduce_proposal_GALLERY_location(
            NXtomoScan(scan="/path/to/PROCESSED_DATA/dataset/my_scan.nx", entry="entry")
        )
        == f"/path/to/{PROPOSAL_GALLERY_DIR_NAME}/dataset"
    )
    assert (
        deduce_proposal_GALLERY_location(
            NXtomoScan(scan="/any/random/path/my_scan.nx", entry="entry")
        )
        == f"/any/random/path/{PROPOSAL_GALLERY_DIR_NAME}"
    )
    assert (
        deduce_proposal_GALLERY_location(
            EDFTomoScan(scan="/path/to/PROCESSED_DATA/dataset/toto")
        )
        == "/path/to/GALLERY/dataset/toto"
    )
    assert (
        deduce_proposal_GALLERY_location(EDFTomoScan(scan="/path/to/dataset/toto"))
        == "/path/to/dataset/toto/GALLERY"
    )



import gc

import pytest
from silx.gui import qt
from silx.gui.utils.testutils import TestCaseQt

from orangecontrib.tomwer.widgets.visualization.NXtomoMetadataViewerOW import (
    NXtomoMetadataViewerOW,
)
from tomwer.tests.utils import skip_gui_test


@pytest.mark.skipif(skip_gui_test(), reason="skip gui test")
class TestNXtomoMetadataViewerOW(TestCaseQt):
    """Test that the NXtomoEditorOW widget work correctly. Processing test are done in the core module. gui test are done in the tomwer.gui.edit module"""

    def setUp(self):
        super().setUp()
        self._window = NXtomoMetadataViewerOW()

    def tearDown(self):
        self._window.setAttribute(qt.Qt.WA_DeleteOnClose)
        self._window.close()
        self._window = None
        gc.collect()

    def test(self):
        self._window.show()
        self.qWaitForWindowExposed(self._window)



from silx.gui import qt
from tomwer.gui.qlefilesystem import QLFileSystem


class HelicalPrepareWeightsDouble(qt.QWidget):
    sigConfigChanged = qt.Signal()
    """emit when the configuration has changed"""

    def __init__(self, parent=None, *args, **kwargs) -> None:
        super().__init__(parent, *args, **kwargs)

        self.setLayout(qt.QGridLayout())
        # TODO: optional outptu file path. Default or set by the user
        self._outputFilePathLabel = qt.QLabel("output file", self)
        self.layout().addWidget(self._outputFilePathLabel, 0, 0, 1, 1)
        self._outputFilePathQLE = QLFileSystem(
            text="{scan_parent_dir_basename}/{scan_dir_name}/map_and_doubleff.hdf5",
            parent=self,
        )
        self._outputFilePathQLE.setToolTip(
            "location of the output file. Which will contain the weight map."
        )
        self.layout().addWidget(self._outputFilePathQLE, 0, 2, 1, 1)

        # transition_width_vertical
        self._verticalTransitionWidthQDSP = qt.QLabel("transition width", self)
        self.layout().addWidget(self._verticalTransitionWidthQDSP, 1, 0, 1, 1)
        self._verticalTransitionWidthQDSP = qt.QDoubleSpinBox(self)
        self._verticalTransitionWidthQDSP.setRange(0.0, 99999999)
        self._verticalTransitionWidthQDSP.setValue(50.0)
        self._verticalTransitionWidthQDSP.setToolTip(
            "the transition width is used to determine how the weights are apodised near the upper and lower border"
        )
        # TODO: improve tooltip of this
        self.layout().addWidget(self._verticalTransitionWidthQDSP, 1, 1, 1, 2)

        # transition_width_vertical
        self._horizontalTransitionWidthQDSP = qt.QLabel("transition width", self)
        self.layout().addWidget(self._horizontalTransitionWidthQDSP, 1, 0, 1, 1)
        self._horizontalTransitionWidthQDSP = qt.QDoubleSpinBox(self)
        self._horizontalTransitionWidthQDSP.setRange(0.0, 99999999)
        self._horizontalTransitionWidthQDSP.setValue(50.0)
        self._horizontalTransitionWidthQDSP.setToolTip(
            "the transition width is used to determine how the weights are apodised near the upper and lower border"
        )
        self.layout().addWidget(self._horizontalTransitionWidthQDSP, 1, 1, 1, 2)

        spacer = qt.QWidget(parent=self)
        spacer.setSizePolicy(qt.QSizePolicy.Expanding, qt.QSizePolicy.Expanding)
        self.layout().addWidget(spacer, 100, 0, 1, 3)

        # connect signal / slot
        self._outputFilePathQLE.editingFinished.connect(self._changed)
        self._verticalTransitionWidthQDSP.valueChanged.connect(self._changed)
        self._horizontalTransitionWidthQDSP.valueChanged.connect(self._changed)

    def _changed(self, *args, **kwargs):
        self.sigConfigChanged.emit()

    def getConfiguration(self) -> dict:
        return {
            "processes_file": self._outputFilePathQLE.text(),
            "transition_width_vertical": self._verticalTransitionWidthQDSP.value(),
            "transition_width_horizontal": self._horizontalTransitionWidthQDSP.value(),
        }

    def setConfiguration(self, config: dict):
        processes_file = config.get("processes_file", None)
        if processes_file is not None:
            self._outputFilePathQLE.setText(processes_file)

        transition_width = config.get("transition_width_vertical", None)
        if transition_width is not None:
            self._verticalTransitionWidthQDSP.setValue(float(transition_width))

        transition_width = config.get("transition_width_horizontal", None)
        if transition_width is not None:
            self._horizontalTransitionWidthQDSP.setValue(float(transition_width))



# coding: utf-8
from __future__ import annotations


from silx.gui import qt

from tomwer.core.process.control.datalistener import DataListener
from tomwer.core.scan.blissscan import BlissScan

from ..processingstack import FIFO, ProcessingThread

try:
    # new HDF5Config class
    from nxtomomill.io.config import TomoHDF5Config as HDF5Config
except ImportError:
    from nxtomomill.io.config import HDF5Config

import logging

_logger = logging.getLogger(__name__)


class DataListenerProcessStack(FIFO, qt.QObject):
    """Stack of file conversion once received by the data-listener
    from a bliss file and a specific entry"""

    def __init__(self, parent=None):
        qt.QObject.__init__(self, parent=parent)
        FIFO.__init__(self, process_id=-1)
        self._results = {}

    def _process(self, data, configuration, callback=None):
        assert isinstance(data, BlissScan)
        assert isinstance(configuration, HDF5Config)
        _logger.info(f"DataListenerProcessStack is processing {data}")
        self._data_currently_computed = data
        self._computationThread.finished.connect(self._end_threaded_computation)

        self._computationThread.init(data=data, configuration=configuration)
        # need to manage connect before starting it because
        self._computationThread.start()

    def _end_threaded_computation(self, callback=None):
        self._computationThread.finished.disconnect(self._end_threaded_computation)
        super()._end_threaded_computation(callback=callback)

    def _create_processing_thread(self, process_id=None) -> qt.QThread:
        thread = _DataListenerConverterThread()
        thread.setParent(self)
        return thread

    def _end_computation(self, data, future_tomo_obj, callback):
        """
        callback when the computation thread is finished

        :param scan: pass if no call to '_computationThread is made'
        """
        if callback is not None:
            callback()
        if data in self._results:
            nx_scan = self._results[data]
            del self._results[data]
        else:
            nx_scan = None
        self.sigComputationEnded.emit(nx_scan, None)
        self._processing = False
        if self.can_process_next():
            self._process_next()

    def register_result(self, bliss_scan, nx_scan):
        self._results[bliss_scan] = nx_scan


class _DataListenerConverterThread(ProcessingThread):
    """
    Thread use to execute the processing of nxtomomill
    """

    def __init__(self):
        ProcessingThread.__init__(self)
        self._scan = None
        self._configuration = None

    def init(self, data, configuration):
        if not isinstance(data, BlissScan):
            raise TypeError(f"Only manage BlissScan. {type(data)} is not managed")
        assert isinstance(configuration, HDF5Config)
        self._scan = data
        self._configuration = configuration

    def run(self):
        self.sigComputationStarted.emit()
        _logger.processStarted(f"Start conversion of bliss scan {self._scan}")

        data_listener = DataListener()
        data_listener.set_configuration(self._configuration or {})
        try:
            scans = data_listener.process_sample_file(
                sample_file=self._scan.master_file,
                entry=self._scan.entry,
                proposal_file=self._scan.proposal_file,
                master_sample_file=self._scan.saving_file,
            )
        except Exception as e:
            _logger.processFailed(
                f"Conversion of bliss scan {self._scan}. Reason is {e}"
            )
            scans = None
        else:
            _logger.processSucceed(f"Conversion of bliss scan {self._scan}.")
        self.parent().register_result(self._scan, scans)



# coding: utf-8
from __future__ import annotations


from ewokscore.task import Task as EwoksTask


class _DiffViewerPlaceHolder(EwoksTask, input_names=("data",)):
    """
    task to compare frames (between one or several dataset)
    """

    def run(self):
        pass



# coding: utf-8
from __future__ import annotations

import logging

_logger = logging.getLogger(__name__)


NABU_CONFIG_FILE_EXTENSION = ".cfg"

NABU_CFG_FILE_FOLDER = "nabu_cfg_files"
# foler where nabu configuraiton will be saved

NABU_TOMWER_SERVING_HATCH = "nabu_tomwer_serving_hatch.h5"
# file used to insure some passing from tomwer to nabu like providing normalization values

try:
    import nabu.app.reconstruct  # noqa: F401
except ImportError:
    try:
        import nabu.resources.cli.reconstruct  # noqa: F401
    except ImportError:
        _logger.warning(
            "Fail to get path to nabu reconstruct main path. Take the most recent path"
        )
        NABU_FULL_FIELD_APP_PATH = "nabu.app.reconstruct"
    else:
        NABU_FULL_FIELD_APP_PATH = "nabu.resources.cli.reconstruct"
else:
    NABU_FULL_FIELD_APP_PATH = "nabu.app.reconstruct"


NABU_CAST_APP_PATH = "nabu.app.cast_volume"

NABU_MULTICOR_PATH = "nabu.app.multicor"



# coding: utf-8
from __future__ import annotations


import functools
import logging
import shutil
import tempfile
import os

from processview.core.manager import DatasetState, ProcessManager
from processview.core.superviseprocess import SuperviseProcess
from silx.gui import qt

from tomwer.core.process.reconstruction.darkref.darkrefscopy import DarkRefsCopy
from tomwer.core.scan.scanbase import TomwerScanBase
from tomwer.core.settings import get_lbsram_path, isOnLbsram
from tomwer.core.utils.lbsram import is_low_on_memory

from ..processingstack import FIFO, ProcessingThread

_logger = logging.getLogger(__name__)


class DarkRefCopyProcessStack(FIFO, qt.QObject):
    """Implementation of the `.AxisTask` but having a stack for treating
    scans and making computation in threads"""

    sigRefSetted = qt.Signal(str)
    """Signal emit when dark or flat are set by a scan. str is the scan identifier"""

    def __init__(self, process_id=None):
        try:
            self._save_dir = tempfile.mkdtemp()
        except Exception as e:
            _logger.warning(
                f"unable to create save dir. Error is {e}. Won't be able to copy any dark or flat"
            )
        qt.QObject.__init__(self)
        FIFO.__init__(self, process_id=process_id)

    @property
    def save_dir(self):
        return self._save_dir

    def __del__(self):
        try:
            shutil.rmtree(self._save_dir)
        except Exception as e:
            _logger.error(e)

    def _process(self, data, configuration: dict, callback=None):
        if not isinstance(data, TomwerScanBase):
            raise TypeError(f"{data} is expected to be an instance of {TomwerScanBase}")
        if not isinstance(configuration, dict):
            raise TypeError(f"{configuration} is expected to be an instance of {dict}")
        ProcessManager().notify_dataset_state(
            dataset=data,
            process=self,
            state=DatasetState.ON_GOING,
        )
        _logger.processStarted(f"dk-flat-copy {data}")
        assert isinstance(data, TomwerScanBase)
        self._data_currently_computed = data
        if isOnLbsram(data) and is_low_on_memory(get_lbsram_path()) is True:
            # if computer is running into low memory on lbsram skip it
            mess = "low memory, skip dk-flat-copy", data.path
            try:
                ProcessManager().notify_dataset_state(
                    dataset=data, process=self, state=DatasetState.SKIPPED
                )
                _logger.processSkipped(mess)
                if callback is not None:
                    callback()
            except Exception as e:
                _logger.error(e)
            try:
                FIFO._end_threaded_computation(self)
            except Exception as e:
                _logger.error(e)
        else:
            self._data_currently_computed = data
            self._computationThread.init(data=data, inputs=configuration)
            # need to manage connect before starting it because
            fct_callback = functools.partial(self._end_threaded_computation, callback)
            self._computationThread.finished.connect(fct_callback)
            self._computationThread.sigRefSetted.connect(self.sigRefSetted)
            self._computationThread.start()

    def _end_computation(self, data, future_tomo_obj, callback):
        """
        callback when the computation thread is finished

        :param scan: pass if no call to '_computationThread is made'
        """
        if not isinstance(data, TomwerScanBase):
            raise TypeError(f"data is {type(data)} when {TomwerScanBase} expected.")
        super()._end_computation(
            data=data, future_tomo_obj=future_tomo_obj, callback=callback
        )

    def _end_threaded_computation(self, callback=None):
        assert self._data_currently_computed is not None
        self._computationThread.finished.disconnect()
        if callback:
            callback()
        FIFO._end_threaded_computation(self)

    def _create_processing_thread(self, process_id=None) -> qt.QThread:
        return _ProcessingThread(process_id=process_id, save_dir=self._save_dir)

    def clear_cache(self):
        """
        remove the file used to cache the reduced darks / flats.
        This can be used in the case it contain unrelevant data. Like frame with another shape...
        """
        cache_file = DarkRefsCopy.get_save_file(self._save_dir)
        if os.path.exists(cache_file):
            os.remove(cache_file)


class _ProcessingThread(ProcessingThread, SuperviseProcess):
    """
    Thread use to execute the processing of the axis position
    """

    sigRefSetted = qt.Signal(str)
    """Signal emit when dark or flat are set by a scan. str is the scan identifier"""

    def __init__(self, save_dir, process_id=None):
        SuperviseProcess.__init__(self, process_id=process_id)
        try:
            ProcessingThread.__init__(self, process_id=process_id)
        except TypeError:
            ProcessingThread.__init__(self)
        self._save_dir = save_dir
        self._data = None
        self._inputs = None

    def init(self, data, inputs):
        self._data = data
        self._inputs = inputs

    def run(self):
        self.sigComputationStarted.emit()
        inputs = self._inputs
        inputs["data"] = self._data
        inputs["save_dir"] = self._save_dir
        inputs["serialize_output_data"] = False
        process = DarkRefsCopyWithSig(
            parent=self.parent(), inputs=inputs, process_id=self.process_id
        )
        process.sigRefSetted.connect(self.sigRefSetted)
        try:
            process.run()
        except Exception as e:
            _logger.warning(e)


class DarkRefsCopyWithSig(qt.QObject, DarkRefsCopy):
    sigRefSetted = qt.Signal(str)
    """Signal emit when dark or flat are set by a scan. str is the scan identifier"""

    def __init__(self, parent=None, *args, **kwargs) -> None:
        qt.QObject.__init__(self, parent)
        DarkRefsCopy.__init__(self, *args, **kwargs)

    def _ref_has_been_set(self, scan):
        self.sigRefSetted.emit(scan.get_identifier().to_str())



import os

import numpy
from nxtomo.application.nxtomo import NXtomo
from scipy.ndimage import shift as shift_scipy
from silx.image.phantomgenerator import PhantomGenerator
from nxtomo.nxobject.nxdetector import ImageKey

from tomwer.core.scan.nxtomoscan import NXtomoScan
from nxtomo.utils.transformation import DetYFlipTransformation, DetZFlipTransformation


def create_scans_z_series(
    output_dir,
    z_positions_m,
    x_positions_m=None,
    pixel_size=10e-6,
    n_proj=20,
    raw_frame_width=100,
    final_frame_width=100,
    shifts=None,
    z_rois=None,
    flip_lr: bool = False,
    flip_ud: bool = False,
):
    """
    :param Optional tuple: shift to be applied on the 2D frame
    """
    if not os.path.exists(output_dir):
        os.makedirs(output_dir)
    frame = (
        PhantomGenerator.get2DPhantomSheppLogan(n=raw_frame_width).astype(numpy.float32)
        * 256.0
    )
    if x_positions_m is None:
        x_positions_m = [None] * len(z_positions_m)
    if shifts is None:
        shifts = [None] * len(z_positions_m)
    if z_rois is None:
        z_rois = [None] * len(z_positions_m)
    scans = []
    for i_frame, (z_pos, x_pos, z_roi, shift) in enumerate(
        zip(z_positions_m, x_positions_m, z_rois, shifts)
    ):
        nx_tomo = NXtomo()
        if z_pos is not None:
            nx_tomo.sample.z_translation = [z_pos] * n_proj
        if x_pos is not None:
            nx_tomo.sample.x_translation = [x_pos] * n_proj
        nx_tomo.sample.rotation_angle = numpy.linspace(
            0, 180, num=n_proj, endpoint=False
        )
        nx_tomo.instrument.detector.image_key_control = [ImageKey.PROJECTION] * n_proj
        if pixel_size is not None:
            if isinstance(pixel_size, (tuple, list)):
                y_pix_size, x_pix_size = pixel_size
            else:
                y_pix_size, x_pix_size = pixel_size, pixel_size
            nx_tomo.instrument.detector.x_pixel_size = x_pix_size
            nx_tomo.instrument.detector.y_pixel_size = y_pix_size
        nx_tomo.instrument.detector.distance = 2.3
        nx_tomo.energy = 19.2
        nx_tomo.instrument.detector.transformations.add_transformation(
            DetYFlipTransformation(flip=flip_ud)
        )
        nx_tomo.instrument.detector.transformations.add_transformation(
            DetZFlipTransformation(flip=flip_lr)
        )

        if shift is not None:
            sub_frame = shift_scipy(
                input=frame.copy(), shift=shift, order=1, mode="constant"
            )[:final_frame_width]
        else:
            sub_frame = frame
        if z_roi is not None:
            roi_start, roi_end = z_roi
            sub_frame = sub_frame[roi_start:roi_end]
        else:
            sub_frame = sub_frame
        if flip_lr:
            sub_frame = numpy.fliplr(sub_frame)
        if flip_ud:
            sub_frame = numpy.flipud(sub_frame)
        nx_tomo.instrument.detector.data = numpy.asarray([sub_frame] * n_proj)

        file_path = os.path.join(output_dir, f"nxtomo_{i_frame}.nx")
        entry = f"entry000{i_frame}"
        nx_tomo.save(file_path=file_path, data_path=entry)
        scans.append(NXtomoScan(scan=file_path, entry=entry))

    return scans



from __future__ import annotations

import logging
import numpy

from silx.gui import qt

from tomwer.core.process.reconstruction.utils.cor import (
    absolute_pos_to_relative,
    relative_pos_to_absolute,
)
from tomwer.gui.utils.buttons import PadlockButton
from tomwer.gui.settings import EDITING_BACKGROUND_COLOR
from tomwer.synctools.axis import QAxisRP

_logger = logging.getLogger(__file__)


class ControlWidget(qt.QWidget):
    """
    Widget to lock, compute or validate the cor position and display it value as absolute or relative
    This is the widget displayed at the bottom of the AxisOW
    """

    sigComputationRequest = qt.Signal()
    """Signal emitted when user request a computation from the settings"""

    sigValidateRequest = qt.Signal()
    """Signal emitted when user validate the current settings"""

    sigLockCORValueChanged = qt.Signal(bool)
    """Signal emitted when the user lock the cor value. Param: True if lock"""

    def __init__(self, parent):
        qt.QWidget.__init__(self, parent)
        self.setLayout(qt.QVBoxLayout())

        # display 'center' information
        self._positionInfo = _PositionInfoWidget(parent=self)
        self.layout().addWidget(self._positionInfo)

        self._buttons = qt.QWidget(parent=self)
        self._buttons.setLayout(qt.QHBoxLayout())
        self.layout().addWidget(self._buttons)

        self._lockBut = PadlockButton(parent=self)
        self._lockBut.setAutoDefault(False)
        self._lockBut.setDefault(False)

        self._buttons.layout().addWidget(self._lockBut)
        self._lockLabel = qt.QLabel("lock cor value", parent=self)
        self._buttons.layout().addWidget(self._lockLabel)

        spacer = qt.QWidget(self)
        spacer.setSizePolicy(qt.QSizePolicy.Expanding, qt.QSizePolicy.Minimum)
        self._buttons.layout().addWidget(spacer)

        self._computeBut = qt.QPushButton("compute", parent=self)
        self._buttons.layout().addWidget(self._computeBut)
        style = qt.QApplication.style()
        applyIcon = style.standardIcon(qt.QStyle.SP_DialogApplyButton)
        self._applyBut = qt.QPushButton(applyIcon, "validate", parent=self)
        self._buttons.layout().addWidget(self._applyBut)
        self.layout().addWidget(self._buttons)

        # set up
        self._positionInfo.setPosition(None, None)

        # make connection
        self._computeBut.pressed.connect(self._needComputation)
        self._applyBut.pressed.connect(self._validate)
        self._lockBut.sigLockChanged.connect(self._lockValueChanged)

    def hideLockButton(self) -> None:
        self._lockLabel.hide()
        self._lockBut.hide()

    def hideApplyButton(self) -> None:
        self._applyBut.hide()

    def _lockValueChanged(self):
        self.sigLockCORValueChanged.emit(self._lockBut.isLocked())
        self._computeBut.setEnabled(not self._lockBut.isLocked())

    def _needComputation(self, *arg, **kwargs):
        """callback when the radio line changed"""
        self.sigComputationRequest.emit()

    def _validate(self):
        self.sigValidateRequest.emit()

    def setPosition(self, relative_cor, abs_cor):
        self._positionInfo.setPosition(relative_cor=relative_cor, abs_cor=abs_cor)

    def _updateAbsolutePosition(self, width):
        self._positionInfo._updateAbsolutePosition(width=width)

    def _updateRelativePosition(self, width):
        self._positionInfo._updateRelativePosition(width=width)

    def isValueLock(self):
        return self._lockBut.isLocked()

    def setValueLock(self, lock):
        self._lockBut.setLock(lock)


class _PositionInfoValidator(qt.QDoubleValidator):
    def validate(self, a0: str, a1: int):
        if "..." in a0 or a0 == "?":
            return (qt.QDoubleValidator.Acceptable, a0, a1)
        else:
            return super().validate(a0, a1)


class _PositionInfoWidget(qt.QWidget):
    """Widget used to display information relative to the current position"""

    sigRelativeValueSet = qt.Signal(float)
    """Emit when the user define manually the relative value"""
    sigAbsolueValueSet = qt.Signal(float)
    """Emit when the user define manually the absolute value"""

    def __init__(self, parent, axis=None):
        self._axis_params = None
        self._defaultBackgroundColor = None
        qt.QWidget.__init__(self, parent)
        self.setLayout(qt.QGridLayout())
        centerLabel = qt.QLabel("center", parent=self)
        centerLabel.setSizePolicy(qt.QSizePolicy.Minimum, qt.QSizePolicy.Minimum)
        font = centerLabel.font()
        font.setBold(True)
        centerLabel.setFont(font)

        self.layout().addWidget(centerLabel, 0, 0, 1, 1)
        self.layout().addWidget(qt.QLabel(" (relative):"), 0, 1, 1, 1)

        self._relativePositionQLE = qt.QLineEdit("", parent=self)
        self._positionValidator = _PositionInfoValidator(self)
        self._relativePositionQLE.setValidator(self._positionValidator)
        self._relativePositionQLE.setSizePolicy(
            qt.QSizePolicy.Expanding, qt.QSizePolicy.Minimum
        )
        self._relativePositionQLE.setStyleSheet("color: red")
        self.layout().addWidget(self._relativePositionQLE, 0, 2, 1, 1)

        centerLabel = qt.QLabel("center", parent=self)
        centerLabel.setSizePolicy(qt.QSizePolicy.Minimum, qt.QSizePolicy.Minimum)
        font = centerLabel.font()
        font.setBold(False)
        centerLabel.setFont(font)

        self.layout().addWidget(centerLabel, 1, 0, 1, 1)
        self.layout().addWidget(qt.QLabel(" (absolute):"), 1, 1, 1, 1)
        self._absolutePositionQLE = qt.QLineEdit("", parent=self)
        self._absolutePositionQLE.setValidator(self._positionValidator)
        self._absolutePositionQLE.setSizePolicy(
            qt.QSizePolicy.Expanding, qt.QSizePolicy.Minimum
        )
        self._absolutePositionQLE.setStyleSheet("color: #ff8c00")
        self.layout().addWidget(self._absolutePositionQLE, 1, 2, 1, 1)

        if axis:
            self.setAxisParams(axis)

        # connect signal / slot
        self._relativePositionQLE.textEdited.connect(
            self._userStartEditingRelativePosition
        )
        self._relativePositionQLE.editingFinished.connect(
            self._userUpdatedRelativePosition
        )
        self._absolutePositionQLE.textEdited.connect(
            self._userStartEditingAbsolutePosition
        )

        self._absolutePositionQLE.editingFinished.connect(
            self._userUpdatedAbsolutePosition
        )

    def _getDefaultBackgroundColor(self):
        if self._defaultBackgroundColor is None:
            self._defaultBackgroundColor = self.palette().color(
                self._absolutePositionQLE.backgroundRole()
            )
        return self._defaultBackgroundColor

    def _userStartEditingRelativePosition(self, *args, **kwargs):
        palette = self.palette()
        palette.setColor(
            self._relativePositionQLE.backgroundRole(), EDITING_BACKGROUND_COLOR
        )
        self._relativePositionQLE.setPalette(palette)

    def _userStartEditingAbsolutePosition(self, *args, **kwargs):
        palette = self.palette()
        palette.setColor(
            self._absolutePositionQLE.backgroundRole(), EDITING_BACKGROUND_COLOR
        )
        self._absolutePositionQLE.setPalette(palette)

    def _userUpdatedRelativePosition(self, *args, **kwargs):
        palette = self.palette()
        palette.setColor(
            self._relativePositionQLE.backgroundRole(),
            self._getDefaultBackgroundColor(),
        )
        self._relativePositionQLE.setPalette(palette)
        if self._relativePositionQLE.text().startswith((".", "?")):
            return
        else:
            value = float(self._relativePositionQLE.text())
            # make sure we only emit the signal if the value changed (and when the Qline has been edited).
            if self._axis_params.relative_cor_value is None or (
                self._axis_params is not None
                and not numpy.isclose(
                    value, self._axis_params.relative_cor_value, atol=1e-3
                )
            ):
                self.sigRelativeValueSet.emit(value)

    def _userUpdatedAbsolutePosition(self, *args, **kwargs):
        palette = self.palette()
        palette.setColor(
            self._absolutePositionQLE.backgroundRole(),
            self._getDefaultBackgroundColor(),
        )
        self._absolutePositionQLE.setPalette(palette)
        if self._absolutePositionQLE.text().startswith((".", "?")):
            return
        else:
            value = float(self._absolutePositionQLE.text())
            # make sure we only emit the signal if the value changed (and when the Qline has been edited).
            if self._axis_params.absolute_cor_value is None or (
                self._axis_params is not None
                and not numpy.isclose(
                    value, self._axis_params.absolute_cor_value, atol=1e-3
                )
            ):
                self.sigAbsolueValueSet.emit(value)

    def setAxisParams(self, axis):
        assert isinstance(axis, QAxisRP)
        if axis == self._axis_params:
            return
        self._axis_params = axis

    def getPosition(self):
        return float(self._relativePositionQLE.text())

    def setPosition(self, relative_cor: float | None, abs_cor: float | None):
        if relative_cor is None:
            self._relativePositionQLE.setText("?")
        elif isinstance(relative_cor, str):
            self._relativePositionQLE.setText(relative_cor)
        else:
            self._relativePositionQLE.setText(f"{relative_cor:.3f}")
        if abs_cor is None:
            self._absolutePositionQLE.setText("?")
        else:
            self._absolutePositionQLE.setText(f"{abs_cor:.3f}")

    def _updateAbsolutePosition(self, width):
        if width is not None:
            try:
                rel_value = float(self._relativePositionQLE.text())
            except Exception:
                return
            else:
                abs_value = relative_pos_to_absolute(
                    relative_pos=float(rel_value), det_width=width
                )
                self._absolutePositionQLE.setText(str(abs_value))

    def _updateRelativePosition(self, width):
        if width is not None:
            try:
                abs_value = float(self._absolutePositionQLE.text())
            except Exception:
                return
            else:
                rel_value = absolute_pos_to_relative(
                    absolute_pos=float(abs_value), det_width=width
                )
                self._relativePositionQLE.setText(str(rel_value))



from __future__ import annotations

import logging

from orangewidget import gui
from orangewidget.settings import Setting
from orangewidget.widget import Input, Output
from silx.gui import qt

import tomwer.core.process.reconstruction.nabu.castvolume
from orangecontrib.tomwer.widgets.utils import WidgetLongProcessing
from tomwer.core.cluster.cluster import SlurmClusterConfiguration
from tomwer.core.futureobject import FutureTomwerObject
from tomwer.core.scan.scanbase import TomwerScanBase
from tomwer.core.volume.volumebase import TomwerVolumeBase
from tomwer.gui.reconstruction.nabu.castvolume import CastVolumeWidget
from tomwer.gui.utils.qt_utils import block_signals

from tomwer.synctools.stacks.reconstruction.castvolume import CastVolumeProcessStack
from tomwer.utils import docstring

from ...orange.managedprocess import SuperviseOW

_logger = logging.getLogger(__name__)


class CastNabuVolumeOW(WidgetLongProcessing, SuperviseOW):
    """
    widget used to cast from 32 bits tiff to 16 bits tiff.

    This is done in a separate process because:

    * this is done in cpu when nabu reconstruct on GPU and this should free sooner GPU resources with the current architecture.
    * this is not included in nabu but also done as post processing.
    * limitation is that not having computed the histogram during volume construction will slow down the cast

    :param parent: the parent widget
    """

    # note of this widget should be the one registered on the documentation
    name = "cast volume"
    id = "orange.widgets.tomwer.reconstruction.CastNabuVolumeOW.CastNabuVolumeOW"
    description = "This widget will allow to cast a nabu volume data type / format to another one."
    icon = "icons/nabu_cast.svg"
    priority = 60
    keywords = [
        "tomography",
        "nabu",
        "reconstruction",
        "volume",
        "cast",
        "tiff",
        "32 bits",
        "16 bits",
        "tif",
    ]

    ewokstaskclass = tomwer.core.process.reconstruction.nabu.castvolume.CastVolumeTask

    want_main_area = True
    resizing_enabled = True

    _ewoks_default_inputs = Setting({"data": None, "cast_volume_params": {}})

    sigScanReady = qt.Signal(TomwerScanBase)
    "Signal emitted when a scan is ended"

    TIMEOUT = 30

    class Inputs:
        data = Input(
            name="data",
            type=TomwerScanBase,
            doc="one scan to be process",
            default=True,
            multiple=False,
        )

        volume = Input(
            name="volume",
            type=TomwerVolumeBase,
            doc="volume to be process",
            default=False,
            multiple=False,
        )
        cluster_in = Input(
            name="cluster_config",
            type=SlurmClusterConfiguration,
            doc="slurm cluster to be used",
            multiple=False,
        )

    class Outputs:
        data = Output(name="data", type=TomwerScanBase, doc="one scan to be process")
        volume = Output(name="volume", type=TomwerVolumeBase, doc="raw volume")
        cast_volume = Output(
            name="cast volume", type=TomwerVolumeBase, doc="cast volume"
        )
        future_tomo_obj = Output(
            name="future_tomo_obj",
            type=FutureTomwerObject,
            doc="future object (process remotely)",
        )

    def __init__(self, parent=None):
        """ """
        SuperviseOW.__init__(self, parent)
        WidgetLongProcessing.__init__(self)
        self._slurmCluster = None
        # processing tool
        self._processingStack = CastVolumeProcessStack(process_id=self.process_id)
        self._window = CastVolumeWidget(parent=self)

        self._layout = gui.vBox(self.mainArea, self.name).layout()
        self._layout.setContentsMargins(0, 0, 0, 0)

        self._layout.addWidget(self._window)

        cast_volume_params = self._ewoks_default_inputs.get("cast_volume_params", {})
        self.setConfiguration(cast_volume_params)

        # connect signal / slot
        self._window.sigConfigChanged.connect(self._updateConfig)
        self._processingStack.sigComputationStarted.connect(self._startProcessing)
        self._processingStack.sigComputationEnded.connect(self._endProcessing)

    def __new__(cls, *args, **kwargs):
        # ensure backward compatibility with 'static_input'
        static_input = kwargs.get("stored_settings", {}).get("static_input", None)
        if static_input not in (None, {}):
            _logger.warning(
                "static_input has been deprecated. Will be replaced by _ewoks_default_inputs in the workflow file. Please save the workflow to apply modifications"
            )
            kwargs["stored_settings"]["_ewoks_default_inputs"] = static_input
        return super().__new__(cls, *args, **kwargs)

    @Inputs.data
    def process_data(self, scan: TomwerScanBase | None):
        self._process(scan=scan, volume=None)

    def _process(self, scan: TomwerScanBase | None, volume: TomwerVolumeBase | None):
        if scan is None and volume is None:
            return
        # check types
        if scan is not None and not isinstance(scan, TomwerScanBase):
            raise TypeError(
                f"scan is expected to be an instance of {TomwerScanBase} not {type(scan)}"
            )
        if volume is not None and not isinstance(volume, TomwerVolumeBase):
            raise TypeError(
                f"volume is expected to be an instance of {TomwerVolumeBase} not {type(volume)}"
            )

        self._processingStack.add(scan or volume, configuration=self.getConfiguration())

    @docstring
    def reprocess(self, dataset):
        if isinstance(dataset, TomwerScanBase):
            self._process(scan=dataset, volume=None)
        elif isinstance(dataset, TomwerVolumeBase):
            self._process(scan=None, volume=dataset)
        else:
            raise TypeError("dataset is expected to be an instance of ''")

    @Inputs.cluster_in
    def setCluster(self, slurm_cluster: SlurmClusterConfiguration | None):
        assert isinstance(
            slurm_cluster, (type(None), SlurmClusterConfiguration)
        ), f"Expect None of SlurmClusterConfiguration. Not {type(slurm_cluster)}"
        self._slurmCluster = slurm_cluster

    @Inputs.volume
    def process_volume(self, volume: TomwerVolumeBase | None):
        self._process(scan=None, volume=volume)

    def getConfiguration(self) -> dict:
        config = self._window.getConfiguration()
        config["cluster_config"] = self._slurmCluster
        return config

    def setConfiguration(self, configuration: dict) -> None:
        with block_signals(self._window):
            self._window.setConfiguration(configuration)

    def _updateConfig(self):
        self._ewoks_default_inputs = {
            "data": None,
            "cast_volume_params": self.getConfiguration(),
        }

    def _endProcessing(self, obj, future_tomo_obj):
        WidgetLongProcessing._endProcessing(self, obj)
        if future_tomo_obj is not None:
            self.Outputs.future_tomo_obj.send(future_tomo_obj)
        if obj is not None:
            if isinstance(obj, TomwerScanBase):
                self.Outputs.data.send(obj)
            elif isinstance(obj, TomwerVolumeBase):
                self.Outputs.volume.send(obj)
            # for now we store a cast_volume to the object but this is not very well design.
            # I guess this will be removed once we move to ewoks or we need to redesign the stack approach
            if obj.cast_volume is not None:
                self.Outputs.cast_volume.send(obj.cast_volume)
